/*! \file ovwater.cpp
 *  \brief Interface between OVWater and Testshell
 */

#include "ovwater.h"

namespace OV
{

/*! \brief Initialization of water module
 *
 *  Stores pointers for all interface variables used between
 *  OVWater and Testshell. Grids and VBO's are also created
 *  by this routine.
 *  
 *  \param camPos Camera position [pointer to string of three floats].
 *  \param camDir Camera direction [pointer to string of three floats].
 *  \param camUp Camera up vector [pointer to string of three floats].
 *  \param time Time variable (seconds).
 *  \param windVelocity Length of vector = wind speed. Direction of vector = wind direction.
 *  \param waterColor Ambient color for the water.
 *  \param sunColor Specular color for the water.
 *  \param lodRes Resolution of mesh [1 3]
 *  \param lodVis Quality level for visualisation [1 3]
 *  \param texCloudID Texture ID for Clouds
 *  \param texSkyID Texture ID for Clouds
 *  \param texGroundID Texture ID for Clouds
 *  \param zoomFactor Increses/Decreses the FOV for zoom effect
 *  \param fovY FOV in y direction
 *  \param fovRation Used in gluProject
 *  \param zNear Near clip plane
 *  \param zFar Far clip plane
 *  \param windowWidth Window width
 *  \param windowHeight Window height
 *  \return void
 */
void OVInitWater(float *camPos, float* camDir, float *camUp, float* time, 
                 float* dayTime, float* waterHorizon, float* windVelocity, 
                 float* waterColor, float* sunColor, int lodRes, int lodVis, 
                 GLuint* texCloudID, GLuint* texSkyID, GLuint* texGroundID,
                 float* zoomFactor, float* fovY, float* fovRatio, float* zNear,
                 float* zFar, int* windowWidth, int* windowHeight)
{
    OVcamPos = camPos;
    OVcamDir = camDir;
    OVcamUp = camUp;

    OVtime = time;
    OVdayTime = dayTime;
    OVwaterHorizon = waterHorizon;
    OVwindVelocity = windVelocity;
    OVwaterColor = waterColor;
    OVsunColor = sunColor;

    OVtexCloudID = texCloudID;
    OVtexSkyID = texSkyID;
    OVtexGroundID = texGroundID;

    //Temp?
    OVwaveVector = OVv2(windVelocity);
    OVamplitude = 1.0;
    OVskyTexUnit = 0;
    OVnoiseTexUnit = 1;
    OVcloudsTexUnit = 2;
    OVgroundTexUnit = 3;

    OValtitude = OVcamPos[1];
  
    OVCalcWaveProperties();
  
    //Create the different resolution meshes
    OVwaterMeshLOW.GenerateGrid(32, 32, -1.0, 1.0, -1.0, 1.0, false);
    OVwaterMeshLOW.CreateVBO();
    OVwaterMeshMEDIUM.GenerateGrid(64, 64, -1.0, 1.0, -1.0, 1.0, false);
    OVwaterMeshMEDIUM.CreateVBO();
    OVwaterMeshHIGH.GenerateGrid(128, 128, -1.0, 1.0, -1.0, 1.0, false);
    OVwaterMeshHIGH.CreateVBO();

    // Call the set functions to ensure all variables have values
    OVSetSunLightUnit();
    OVSetSkyTexUnit();
    OVSetResolutionLOD(lodRes);
    OVSetVisualizationLOD(lodVis);
    OVSetWaveAmpFactor(); //This was supposed to enables scaling of waves, not sure if it ever got implemented though! 

    OVmaxDisplacement = 1.42*(OVwaveProp1[0]+OVwaveProp2[0]+OVwaveProp3[0]+
                            OVwaveProp4[0]+OVwaveProp5[0]+OVwaveProp6[0]);
    float basePlaneY = 0.0;    
   
    OVprojector->InitProjector(camPos, camDir, camUp, &OVmaxDisplacement, 
        basePlaneY, zoomFactor, fovY, fovRatio, zNear, zFar, 
        windowWidth, windowHeight); 

    //Load textures
    OVnoiseTexture.Load("ovwater/ovtextures/noise.tga",true);
    OVskyTexture.Load("ovwater/ovtextures/sky.tga");

    //Create and compile shaders
    OVLoadShader(OVwaterShaderLOW, "ovwater/ovshaders/ovlod1");
    OVLoadShader(OVwaterShaderMEDIUM, "ovwater/ovshaders/ovlod2");
    OVLoadShader(OVwaterShaderHIGH, "ovwater/ovshaders/ovlod3");
}

/*! \brief Loads water shaders
 *
 *  \return void.
 */
void OVLoadShader(OVShader &shader, char* fileName)
{
    shader.Compile(fileName);
    shader.AddUniform(new OVUniVari("texNoise", &OVnoiseTexUnit));
    shader.AddUniform(new OVUniVarf("t", OVtime));
    shader.AddUniform(new OVUniVarf("dayTime", OVdayTime));    
    shader.AddUniform(new OVUniVarf("waterHorizon", OVwaterHorizon));
    shader.AddUniform(new OVUniVarf("A", &OVamplitude));
    shader.AddUniform(new OVUniVarf("altitude", &OValtitude));
    shader.AddUniform(new OVUniVar2fv("k", OVwaveVector));
    shader.AddUniform(new OVUniVar2fv("k1", OVwaveVec1));
    shader.AddUniform(new OVUniVar2fv("k2", OVwaveVec2));
    shader.AddUniform(new OVUniVar2fv("k3", OVwaveVec3));
    shader.AddUniform(new OVUniVar2fv("k4", OVwaveVec4));
    shader.AddUniform(new OVUniVar2fv("k5", OVwaveVec5));
    shader.AddUniform(new OVUniVar2fv("k6", OVwaveVec6));
    shader.AddUniform(new OVUniVar2fv("k7", OVwaveVec7));
    shader.AddUniform(new OVUniVar2fv("k8", OVwaveVec8));
    shader.AddUniform(new OVUniVar4fv("wav1", OVwaveProp1));
    shader.AddUniform(new OVUniVar4fv("wav2", OVwaveProp2));
    shader.AddUniform(new OVUniVar4fv("wav3", OVwaveProp3));
    shader.AddUniform(new OVUniVar4fv("wav4", OVwaveProp4));
    shader.AddUniform(new OVUniVar4fv("wav5", OVwaveProp5));
    shader.AddUniform(new OVUniVar4fv("wav6", OVwaveProp6));
    shader.AddUniform(new OVUniVar4fv("wav7", OVwaveProp7));
    shader.AddUniform(new OVUniVar4fv("wav8", OVwaveProp8));
    shader.AddUniform(new OVUniVar16fm("rangeMatrix", OVrangeMatrix));
    shader.AddUniform(new OVUniVar16fm("projectorModelViewProjectionMatrixInverse", OVprojectorModelViewProjectionMatrixInverse));

    shader.AddUniform(new OVUniVari("texClouds", &OVcloudsTexUnit));
    shader.AddUniform(new OVUniVari("texSky", &OVskyTexUnit));
    shader.AddUniform(new OVUniVari("texGround", &OVgroundTexUnit));
    shader.AddUniform(new OVUniVari("sunLightUnit", &OVsunLightUnit));

    shader.Activate();
    shader.UpdateUniformsLocation();
    shader.Deactivate();
}

/*! \brief Rotates a 2D vector by angle 
 *  \param lengthScale scales the length of the vector.
 *  \param k The vector which should be rotated.
 *  \param angle The desired angle of rotation.
 *  \param newK The new vector after rotation.
 *  \return void.
 */
void OVRotate(float lengthScale, OVv2 k, float angle, OVv2& newK)
{
    float x = lengthScale * k[0];
    float y = lengthScale * k[1];
    
    newK[0] = cos(angle)*x + sin(angle)*y;
    newK[1] = cos(angle)*y - sin(angle)*x;
}

/*! \brief Calculates the waves amplitude, length, steepness, and speed 
 *
 *  \return void.
 */ 
void OVCalcWaveProperties()
{
    //Calculate wave properties
    float scl = 5.0;
    OVRotate(scl                 , OVwaveVector,  0.00, OVwaveVec1);
    OVRotate(scl/sqrt(2.f) + 0.04, OVwaveVector, -0.09, OVwaveVec2);
    OVRotate(scl/sqrt(3.f) + 0.09, OVwaveVector,  0.17, OVwaveVec3);
    OVRotate(scl/sqrt(4.f) - 0.04, OVwaveVector, -0.03, OVwaveVec4);
    OVRotate(scl/sqrt(5.f)       , OVwaveVector,  0.21, OVwaveVec5);
    OVRotate(scl/sqrt(6.f) + 0.04, OVwaveVector, -0.19, OVwaveVec6);

    OVCalcWaveProp(1, OVwaveProp1, 1.01, OVwaveVec1);
    OVCalcWaveProp(2, OVwaveProp2, 0.94, OVwaveVec2);
    OVCalcWaveProp(3, OVwaveProp3, 0.87, OVwaveVec3);
    OVCalcWaveProp(4, OVwaveProp4, 0.79, OVwaveVec4); 
    OVCalcWaveProp(5, OVwaveProp5, 0.69, OVwaveVec5);
    OVCalcWaveProp(6, OVwaveProp6, 0.55, OVwaveVec6); 
}

/*! \brief Calculates the waves amplitude, length, steepness, and speed 
 *  \param wave The wave index.
 *  \param waveProp Vector (4 elements) where the wave properties are stored.
 *  \param scaleFactor Constant used for scaling the amplitude.
 *  \param waveVec Vector conOVwaveProp6taining direction and size of wave.
 *  \return void.
 */ 
void OVCalcWaveProp(int wave, OVv4& waveProp, float scaleFactor, OVv2 waveVec)
{

    float numWaves = 6.0;
    float lengthWaveVec = sqrt(waveVec[0]*waveVec[0] + waveVec[1]*waveVec[1]);
    float Q = 0.5 + lengthWaveVec*(float)wave/700.0;
    float g = 9.82;
    float pi = 3.141592; 

    //if(lengthWaveVec > 15) lengthWaveVec = 15.0;
    //if(lengthWaveVec < 0.3) lengthWaveVec = 0.3;

    if(fabs(OValtitude) < 10)
    {
        if(OValtitude >= 0)
        {
            OValtitude = 10.0;
        }
        else
        {
            OValtitude = -10.0;
        }        
    }    

    float amp = 0.5*scaleFactor*OVamplitude / sqrt(fabs(OValtitude/4)); 
    float freq = 1.0/lengthWaveVec; 
      
    waveProp[0] = amp;
    waveProp[1] = freq;
    waveProp[2] = Q/(freq*amp*numWaves);
    waveProp[3] = 0.05*sqrt(g*2.0*pi/freq);  // borde inte vara någon konstant här!!! men såg bättre ut
}

/*! \brief Draws the water, requires previous call to OVInitWater
 *  \return void.
 */
void OVDrawWater()
{
    OValtitude = OVcamPos[1];
      
    // water material properties
    // SKA VI? spara undan material på något sätt...
    glDisable(GL_CULL_FACE);
    GLfloat mat_shininess[]={ 32.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT, OVwaterColor);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, OVwaterColor);
    glMaterialfv(GL_FRONT, GL_SPECULAR, OVsunColor);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess );
   
    //Calculate the wave vector and wave amplitude
    float windSpeed = sqrt(OVwindVelocity[0]*OVwindVelocity[0] +
                  OVwindVelocity[1]*OVwindVelocity[1]);
    
    // Limit and scale the size of waves
    if(windSpeed > 5.0)
    {
        OVwaveVector[0] = 0.15*OVwindVelocity[0];
        OVwaveVector[1] = 0.15*OVwindVelocity[1];
    }
    else
    {
        OVwaveVector[0] = OVwindVelocity[0]/windSpeed*0.75;
        OVwaveVector[1] = OVwindVelocity[1]/windSpeed*0.75;
    }

    OVamplitude = OVscaleAmplitude*sqrt(OVwaveVector[0]*OVwaveVector[0] +
                  OVwaveVector[1]*OVwaveVector[1]);

    OVCalcWaveProperties(); 
    
    //Calculate the maximum vertex displacement (both in y- and xz-directions)
    //Used by the projector
    OVmaxDisplacement = 1.42*(OVwaveProp1[0]+OVwaveProp2[0]+OVwaveProp3[0]
                        +OVwaveProp4[0]+OVwaveProp5[0]+OVwaveProp6[0]);
 
    //Set the projector    
    OVprojector->SetUpProjector();

    //Get the scaling matrix used by the shader
    OVrangeMatrix = OVprojector->GetMRange();  

    //Make sure the correct shader is running
    OVactiveShader->Activate();    
    
    //First save the camera's modelviewprojection matrix
    //then set the modelview matrix according to the projector
    
    //Push camera matrices
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    
    //Set up for projection
    OVprojector->Perspective();
    OVprojector->Look();        

    //Store projection matrices
    OVm4 projModelviewMatrix;
    glGetFloatv(GL_MODELVIEW_MATRIX, projModelviewMatrix);
    OVm4 projPerspectiveMatrix;
    glGetFloatv(GL_PROJECTION_MATRIX, projPerspectiveMatrix);
    OVprojectorModelViewProjectionMatrixInverse = projModelviewMatrix * projPerspectiveMatrix;
    OVprojectorModelViewProjectionMatrixInverse = OVprojectorModelViewProjectionMatrixInverse.GetInverse();
    
    //Pop camera matrices
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
           
    //Setup textures
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, *OVtexSkyID);
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glBindTexture(GL_TEXTURE_2D, *OVtexCloudID);
    glActiveTextureARB(GL_TEXTURE3_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, *OVtexGroundID);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER ,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER ,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_2D, OVnoiseTexture.texID);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER ,GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER ,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    //Render
    OVactiveShader->UpdateUniforms();
    
    if(OVprojector->Visible())
    { 
        OVactiveMesh->DrawVBO();
    }

    glMatrixMode(GL_MODELVIEW);

    //Remove the projector's modelview matrix
    //and go back to the camera's
    //Remove the projector's perspective matrix
    //and go back to the camera's
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);        
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    
    //Deactivate textures
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE3_ARB);
    glDisable(GL_TEXTURE_2D);

    //Deactivate shader
    OVactiveShader->Deactivate();
}

/*! \brief Access the water height.
 *
 *  Returns the height of the water at specified point.
 *  The height will only be an taylor approximation to
 *  the first order since it is almost impossible to get
 *  an exact height due to the vertex displacement in 
 *  the xz-plane.
 *  
 *  \param x x-coordinate for height lookup.
 *  \param z z-coordinate for height lookup.
 *  \return Height value.
 */
float OVGetHeight(float x,float z)
{
    int numWaves = 6;

    //Create an array of all waveprops and wavevectors
    OVv4* waveProps[8] = {&OVwaveProp1, &OVwaveProp2, &OVwaveProp3, 
                          &OVwaveProp4, &OVwaveProp5, &OVwaveProp6, 
                          &OVwaveProp7, &OVwaveProp8};

    OVv2* waveVecs[8] = {&OVwaveVec1, &OVwaveVec2, &OVwaveVec3, &OVwaveVec4,
                         &OVwaveVec5, &OVwaveVec6, &OVwaveVec7, &OVwaveVec8};

    //Calculate the height based on the approximated xz coordinates
    float height = 0.f;
    for(int i = 0; i < numWaves; i++)
    {
        OVv2 d = *(waveVecs[i]);
        float lengthD = sqrt(d[0]*d[0]+d[1]*d[1]);
        float dHat[2] = {d[0]/lengthD, d[1]/lengthD};

        float w = (*(waveProps[i]))[1];
        float A = (*(waveProps[i]))[0];
        float phi = (*(waveProps[i]))[3];

        height += A*sin(w*(dHat[0]*x+dHat[1]*z)-phi*(*OVtime));
    }

    return height;
}

/*! \brief Sets OpenGL light unit used for the sun (eg GL_LIGHT0).
 *  \param lightUnitID ID of the light unit. 
 *  \return void
 */
void OVSetSunLightUnit(int lightUnitID)
{
    int unit = lightUnitID > 7 ? 7 :
                lightUnitID < 0 ? 0 : lightUnitID;
    
    OVsunLightUnit = unit;
}

/*! \brief Sets OpenGL texture unit used for the sky.
 *  \param texUnitID ID of the texture unit. 
 *  \return void
 */
void OVSetSkyTexUnit(int texUnitID) 
{
    int maxTexUnits;
    glGetIntegerv(GL_MAX_TEXTURE_UNITS, &maxTexUnits);
    maxTexUnits -= 1;
    int unit = texUnitID > maxTexUnits ? maxTexUnits :
                texUnitID < 0 ? 0 : texUnitID;
    
    OVskyTexUnit = unit;
}

/*! \brief Sets resolution level of detail.
 *  \param lodRes Value of 1, 2 or 3. 
 *  \return void
 */
void OVSetResolutionLOD(int lodRes)
{
    // Update corresponding variable!    
    
    int lod = lodRes > 3 ? 3 :
                lodRes < 1 ? 1 : lodRes;
    
    OVlodRes = lod;
    
    switch(lod)
    {
    case 1:
        OVactiveMesh = &OVwaterMeshLOW;
        break;
    case 2:
        OVactiveMesh = &OVwaterMeshMEDIUM;
        break;
    case 3:
        OVactiveMesh = &OVwaterMeshHIGH;
        break;
    default:
        //Default to the medium lod (should not be possible)
        OVactiveMesh = &OVwaterMeshMEDIUM;
        break;
    }  
     
}

/*! \brief Sets visualization level of detail.
 *  \param lodVis Value of 1, 2 or 3. 
 *  \return void
 */
void OVSetVisualizationLOD(int lodVis)
{
    //Make sure a valid lod is choosen
    int lod = lodVis > 3 ? 3 :
                lodVis < 1 ? 1 : lodVis;
    
    OVlodRes = lod;

    switch(lod)
    {
    case 1:
        OVactiveShader = &OVwaterShaderLOW;
        break;
    case 2:
        OVactiveShader = &OVwaterShaderMEDIUM;
        break;
    case 3:
        OVactiveShader = &OVwaterShaderHIGH;
        break;
    default:
        //Default to the medium lod (should not be possible)
        OVactiveShader = &OVwaterShaderMEDIUM;
        break;
    }        
}

/*! \brief Sets additional amplification of wave amplitude.
 *  \param factor Factor to multiply wave amplitude. 
 *  \return void
 */
void OVSetWaveAmpFactor(float factor)
{
    OVscaleAmplitude = factor;
//    OVCalcWaveProperties();
}

/*! \brief Draws the projector frustum.
*   \return void
*/
void OVDrawProjectorFrustum()
{
    OVprojector->DrawFrustum();
}

}  // End of OV
