/*! \file ovlod3.frag
 *  \brief LOD 3 fragment shader
 */

uniform sampler2D texSky;
uniform sampler2D texNoise;
uniform float A;
uniform float t;
uniform float altitude;
uniform float waterHorizon;
uniform float dayTime;

// rotated wavevectors
uniform vec2 k;
uniform vec2 k1;
uniform vec2 k2;
uniform vec2 k3;
uniform vec2 k4;
uniform vec2 k5;
uniform vec2 k6;

// waveProps
uniform vec4 wav1;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav2;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav3;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav4;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav5;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav6;  //Wave amplitude, freq, steepness, and speed

varying vec4 fragPos;

varying vec2 texCoord1;
varying vec2 texCoord2;
varying vec2 texCoord3;

varying float ground_height;
varying float antiAliasHorizon;
varying float antiAliasAltitude;

float antiAliasNormals;

/*! \brief Computes the normal for gerstner waves
 *  
 *  \param w the spatial frequency of the wave
 *  \param A the amplitude of the wave
 *  \param D the direction (i.e. wave vector) of the wave
 *  \param Q the steepness of the wave
 *  \param phi the temporal frequency of the wave
 *  \return the normal
 */
vec3 GetNormal(float w, float A, vec2 D, float Q, float phi)
{
    vec2 Dn = normalize(D);
    float trigArg = w*dot(Dn,fragPos.xz) - phi*t;
    float wA = w*A;
    float normalXZ = wA*cos(trigArg);
    return vec3(D.x*normalXZ, Q*wA*sin(trigArg), D.y*normalXZ);
}

void main()
{
    // Noise nr 1
    float noise1 = texture2D(texNoise, texCoord1);

    vec4 fragPos_C = gl_ModelViewMatrix * fragPos;  
    antiAliasNormals = 1.0/max(1.0, 0.015*length(fragPos_C)/sqrt(altitude)); 
    
    // Noise nr 2
    float noise2 = texture2D(texNoise, vec2(texCoord2.x , texCoord2.y ) );

    vec3 norm = vec3(0.0, 1.0, 0.0);
    
    // Noise nr 3
    float noise3 = antiAliasNormals*
                    texture2D(texNoise, vec2(texCoord3.x, texCoord3.y));

    float noise = mix(noise1, 
                      mix(noise2, noise3, 0.25), 
                      0.5*antiAliasAltitude) - 0.5;

    vec2 kn = normalize(k);
    vec3 noiseNormal = vec3(kn.x*noise, 1.0, kn.y*noise);

    // The normal used for lighting calculations (gerstner + noise)
    vec3 normalN = normalize(mix(normalize(noiseNormal), 
                                 normalize(norm), antiAliasHorizon));

    // Sun reflections (*_C for camera space)  
    vec4 cameraPos_C = vec4(0.0, 0.0, 0.0, 1.0);
    vec3 lightDir_C = -gl_LightSource[0].spotDirection;
    vec3 normalDir_C = normalize(gl_NormalMatrix * normalN);
    vec3 reflectionDir_C = -normalize(reflect(lightDir_C, normalDir_C));

    // Sky reflections
    vec4 cameraPos = gl_ModelViewMatrixInverse[3];
    vec3 cameraDir = normalize(cameraPos.xyz - fragPos.xyz);
    vec3 skyReflectionDir = normalize(reflect(cameraDir, normalN));

    //Adjust the height a little (compensating for the horizon being too close).
    cameraPos.y = pow(cameraPos.y,0.8);
    cameraPos_C = gl_ModelViewMatrix * cameraPos;
    vec3 cameraDir_C = normalize(cameraPos_C.xyz - fragPos_C.xyz);

    // Sky mapping
    float height = 1.0;
    vec2 texCoordSky = vec2(dayTime, 1.0 + skyReflectionDir.y/height);

    // Clamp takes away the strong white parts of the sky
    vec4 skyColor = texture2D(texSky, clamp(texCoordSky, vec2(0.0), vec2(0.9))); 

    float NdotL = max(dot(normalDir_C,lightDir_C),0.0);
    float RdotL = dot(reflectionDir_C,lightDir_C);
    float RdotC = max(dot(reflectionDir_C,cameraDir_C),0.0);
    float NdotC = dot(normalDir_C, cameraDir_C);

    // Approximate the fresnel factor
    float F = max(1.0/pow(1.0 + NdotC/5.0, 8.0), 0.1);

    // Ambient light
    vec4 ambColor = gl_FrontMaterial.ambient * gl_LightSource[0].ambient;
    ambColor +=	gl_LightModel.ambient * gl_FrontMaterial.ambient;
    
    // Diffuse reflection
    vec4 diffuse = NdotL*1.2*sqrt(1.0-pow(1.0-1.0*dayTime/0.24,2.0)) *
                    gl_FrontMaterial.diffuse;

    // Specular reflection.
    vec4 specular = gl_LightSource[0].specular * 
                     pow(RdotC, gl_FrontMaterial.shininess*antiAliasAltitude);

    // Mix specular and diffuse reflections depending on altitude and wind.
    vec4 color = mix(diffuse, specular/length(k), 0.6);
    
    // Mix in the sky color
    color += mix(skyColor, color, 0.8); 

    // Mix away patterns att high altitude
    color = mix(ambColor, color, antiAliasAltitude); 

    // Scale the color using the fresnel factor, depends on the time of day
    float ambColorAdjust = 1.0 - abs(dayTime - 0.5);
    color = mix(ambColorAdjust*ambColor, color, F);

    // White around islands
    color += mix(ambColor, 1.0, noise1)*ground_height*noise3;

    //Blur the horizon
    vec2 horizonCoords = vec2(dayTime, 0.8*pow(antiAliasHorizon,5.0));
    vec4 horizonColor = texture2D(texSky, clamp(horizonCoords,0.0,0.9));
    color = mix(color, horizonColor, pow(antiAliasHorizon,3.0));
    
    gl_FragColor = color;
}
