/*! \file ovlod3.vert
 *  \brief LOD 3 vertex shader
 */

uniform sampler2D texGround;

uniform float t;
uniform float A;
uniform float altitude;
uniform float waterHorizon;

// rotated wavevectors
uniform vec2 k1;
uniform vec2 k2;
uniform vec2 k3;
uniform vec2 k4;
uniform vec2 k5;
uniform vec2 k6;

// waveProps
uniform vec4 wav1;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav2;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav3;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav4;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav5;  //Wave amplitude, freq, steepness, and speed
uniform vec4 wav6;  //Wave amplitude, freq, steepness, and speed

uniform mat4 rangeMatrix; //range matrix used for optimization during projection
uniform mat4 projectorModelViewProjectionMatrixInverse; //transforms from world space to image (projector) space

varying vec4 Vertex;
varying vec4 fragPos;
varying float antiAliasHorizon;
varying float antiAliasAltitude;

varying vec2 texCoord1;
varying vec2 texCoord2;
varying vec2 texCoord3;

vec4 pInt;
vec4 color;

/*! \brief Computes texture coordinates
 *  
 *  Calculates texture coordinates for the noise texture for waves moving
 *  with wave vector k.
 *  
 * \param The wave vector for the wave (i.e direction and speed of the wave)
 * \return the calculated texture  coordinates
 */
vec2 GetTexCoord(vec2 k)
{
    return 0.05*(k*dot(k, (Vertex.xz + fragPos.xz)/2.0) + 
            0.1*vec2(k.y,-k.x)*dot(vec2(k.y,-k.x), (Vertex.xz + fragPos.xz)/2.0));
}

void main()
{
    //Colors
    vec4 cBlue = vec4(0.05,0.2,0.6, 1.0);
    vec4 cRed  = vec4(0.81,0.0,0.0, 0.0);

    //Project extremes to form line section
    vec4 pMin = gl_Vertex.xzyw;
         pMin.z = 0.0;
         //make the coordinate range [0,1] to be compatible with the range matrix
         pMin.x= (pMin.x+1.0)/2.0;
         pMin.y= (pMin.y+1.0)/2.0;
         pMin = projectorModelViewProjectionMatrixInverse * rangeMatrix *  pMin;
         pMin = pMin / pMin.w;
    vec4 pMax = gl_Vertex.xzyw;
         pMax.z = 1.0;
         //make the coordinate range [0,1] to be compatible with the range matrix
         pMax.x= (pMax.x+1.0)/2.0;
         pMax.y= (pMax.y+1.0)/2.0;
         pMax = projectorModelViewProjectionMatrixInverse * rangeMatrix *  pMax;
         pMax = pMax / pMax.w;
    /* 
     * 1: pMax.y > p.Min.y 
     *    No intersection (pMin.y > 0 always)
     * 2: pMin.y > p.Max.y > 0
     *    No intersection
     * 3: pMin.y > 0 > p.Max.y
     *    INTERSECTION
     * Solve with step functions to avoid if()'s in shader!
     */
    float isInt = abs(step(pMax.y, 0.0) - step(pMin.y, 0.0)); //intersection?

    float k = (0.0 - pMax.y) / (pMin.y - pMax.y); //relative above/below
    vec4 intersection;
    intersection.x = pMax.x + k*(pMin.x - pMax.x); //x
    intersection.y = 0.0;                         //y
    intersection.z = pMax.z + k*(pMin.z - pMax.z); //z
    intersection.w = 1.0;                          //1.0

    vec4 dump = pMax + normalize(pMax-pMin);

    pInt = mix(dump,intersection,isInt); //set according to a
    Vertex = pInt;

    vec4 cameraPos = gl_ModelViewMatrixInverse[3];
    float vertDistance = length(cameraPos.xyz-Vertex.xyz);
    antiAliasHorizon = vertDistance/waterHorizon;
    antiAliasAltitude = 1.0 / max( sqrt((altitude - 500.0)/500.0), 1.0);

    fragPos = Vertex;
 
    gl_Position = gl_ModelViewProjectionMatrix * fragPos;

    texCoord1 = (GetTexCoord( 0.23 * normalize(k1)/sqrt(length(k1)) ) - 0.01*normalize(k1)*t);
    texCoord2 = (GetTexCoord( 0.67 * normalize(k4)/sqrt(length(k4)) ) - 0.08*normalize(k4)*t);
    texCoord3 = (GetTexCoord( 1.67 * normalize(k2)/sqrt(length(k2)) ) - 0.50*normalize(k2)*t);  
}








