/*! \file ovalgebra.cpp
 *  \brief Source file for OV matrix containers.
 *
 *  Contains the definitions of the OV matrix classes.
 */

#include "ovalgebra.h"

namespace OV
{

//=============================================================//
//                                                             //
//                       3x3 Matrix                            //
//                                                             //
//=============================================================//

/*! \brief Constructor from 9 elements.
 *  \param eXX = eRC, element (R,C) in matrix. */
OVm3::OVm3(float e00,float e01,float e02,
		   float e10,float e11,float e12,
           float e20,float e21,float e22) {
	float a[16] = {e00,e01,e02,
	               e10,e11,e12,
	               e20,e21,e22};
	this->Set(a);
}

/*! \brief Sets matrix elements.
 *  \param a Array of 9 floats. */
void OVm3::Set(float *a) {
	*this = a;
}

/*! \brief Access operator for entire row of matrix.
 *  \param index Row index.
 *  \return Returns indexed row as OV3f vector. */
OVv3 OVm3::GetRow(int index) {
    index = (index < 0) ? 0 : ((index > 2) ? 2 : index);
    float row[3] = {e[3*index+0], e[3*index+1], e[3*index+2]};
	return OVv3(row);
}

/*! \brief Access operator for entire column of matrix.
 *  \param index Column index.
 *  \return Returns indexed column as OV3f vector. */
OVv3 OVm3::GetCol(int index) {
    index = (index < 0) ? 0 : ((index > 2) ? 2 : index);
	float col[3] = {e[0+index], e[3+index], e[6+index]};
	return OVv3(col);
}

/*! \brief Sets one row of matrix elements.
 *  \param index Row index. 
 *  \param v Contains the elements to which the row should be set
 */
void OVm3::SetRow(int index,float *v) {
    index = (index < 0) ? 0 : ((index > 2) ? 2 : index);
	e[3*index  ] = v[0];
	e[3*index+1] = v[1];
	e[3*index+2] = v[2];
}

/*! \brief Sets one column of matrix elements.
 *  \param index Column index. 
 *  \param v Contains the elements to which the column should be set
 */
void OVm3::SetCol(int index,float *v) {
    index = (index < 0) ? 0 : ((index > 2) ? 2 : index);
	e[   index] = v[0];
	e[3 +index] = v[1];
	e[6 +index] = v[2];
}

/*! \brief Clears matrix to 0. */
void OVm3::Clear() {
	for(uint i=0;i<9;i++)
		e[i] = 0;
}

/*! \brief Sets matrix to identity. */
void OVm3::SetToIdentityMatrix() {
	Clear();
	e[0] = 1;
	e[4] = 1;
	e[8] = 1;
}

/*! \brief Prints matrix to commandline. */
void OVm3::Print() {
	std::printf("%f %f %f \n%f %f %f \n %f %f %f\n\n",
		e[0],e[1],e[2],e[3],e[4],e[5],e[6],e[7],e[8]);
}

/*! \brief << standard C ostream operator.
 *  \param out Reference to active ostream.
 *  \param m Reference to matrix.
 *  \return Returns the ostream reference in parameter out. */
std::ostream& operator<<(std::ostream &out, const OVm3 &m)
{
	out << m[0] << " " << m[1] << " " <<  m[2] << std::endl;
	out << m[3] << " " << m[4] << " " <<  m[5] << std::endl;
	out << m[6] << " " << m[7] << " " <<  m[8];
	return out;
}

/*! \brief Operator for multiplication with scalar.
 *  \param f Constant scalar.
 *  \return Returns result of multiplication as OVm3 matrix. */
OVm3 OVm3::operator*(const float f) {
	OVm3 m;
	for(int i=0;i<9;i++)
		m[i] = e[i]*f;
	return m;
}

/*! \brief Operator for multiplication with vector.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OV3f.
 *  \param v Constant reference to vector.
 *  \return Returns result of multiplication as OV3f vector. */
OVv3 OVm3::operator*(const OVv3 &v) {
	OVv3 vnew;
	vnew[0] = e[0]*v[0] + e[1]*v[1] + e[2]*v[2];
	vnew[1] = e[3]*v[0] + e[4]*v[1] + e[5]*v[2];
	vnew[2] = e[6]*v[0] + e[7]*v[1] + e[8]*v[2];
	return vnew;
}

/*! \brief Operator for multiplication with 3x3 matrix.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OVm3.
 *  \param m Constant reference to matrix.
 *  \return Returns result of multiplication as OVm3 matrix. */
OVm3 OVm3::operator*(const OVm3 &m) {
	OVm3 a;
	a[0] = e[0]*m[0] + e[1]*m[3] + e[2]*m[6];
	a[1] = e[0]*m[1] + e[1]*m[4] + e[2]*m[7];
	a[2] = e[0]*m[2] + e[1]*m[5] + e[2]*m[8];
	a[3] = e[3]*m[0] + e[4]*m[3] + e[5]*m[6];
	a[4] = e[3]*m[1] + e[4]*m[4] + e[5]*m[7];
	a[5] = e[3]*m[2] + e[4]*m[5] + e[5]*m[8];
	a[6] = e[6]*m[0] + e[7]*m[3] + e[8]*m[6];
	a[7] = e[6]*m[1] + e[7]*m[4] + e[8]*m[7];
	a[8] = e[6]*m[2] + e[7]*m[5] + e[8]*m[8];
	return a;
}

/*! \brief = operator.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OVm3.
 *  \param a Constant reference to matrix. */
void OVm3::operator=(const float *a) {
	for(uint i=0;i<9;i++)
		e[i] = a[i];
}

/*! \brief Determinant of any 2x2 matrix.
 *  \param eXX eXX = eRC, element (R,C) in matrix.
 *  \return Scalar value of determinant. */
float OVm3::Det2x2(float &e00,float &e01,float &e10,float &e11) {
	return (e00*e11-e01*e10);
}

/*! \brief Determinant of any 3x3 matrix.
 *  \param eXX eXX = eRC, element (R,C) in matrix.
 *  \return Scalar value of determinant. */
float OVm3::Det3x3(float &e00,float &e01,float &e02,
		                  float &e10,float &e11,float &e12,
                          float &e20,float &e21,float &e22) {
	return (e00*(e11*e22 - e12*e21) +
			e01*(e12*e20 - e10*e22) +
			e02*(e10*e21 - e11*e20));
}

/*! \brief Determinant of object matrix.
 *  \return Scalar value of determinant. */
float OVm3::GetDeterminant() {
	return Det3x3(e[0],e[1],e[2],
				  e[3],e[4],e[5],
				  e[6],e[7],e[8]);
}

/*! \brief Transpose of object matrix.
 *  \return Transpose OVm3 matrix. */
OVm3 OVm3::GetTranspose() {
	OVm3 m;
	m[0] = e[0];
	m[1] = e[3];
	m[2] = e[6];
	m[3] = e[1];
	m[4] = e[4];
	m[5] = e[7];
	m[6] = e[2];
	m[7] = e[5];
	m[8] = e[8];
	return m;
}

/*! \brief Inverse of object matrix.
 *  Using adjoint transpose matrices to calculate inverse.
 *  \return Inverse OVm3 matrix. */
OVm3 OVm3::GetInverse() {
	// Set up transpose
   float a1, a2, a3, b1, b2, b3, c1, c2, c3;
   a1 = e[0]; b1 = e[3]; c1 = e[6];
   a2 = e[1]; b2 = e[4]; c2 = e[7];
   a3 = e[2]; b3 = e[5]; c3 = e[8];
   float det = Det3x3(a1, a2, a3, b1, b2, b3, c1, c2, c3);
   // Check if singular matrix
   if(fabs(det) < 0.0000001) {
      std::printf("(OVm3.inverse): Singular matrix, unable to calculate inverse\n");
      return OVm3();
   }
   float invdet = 1.0f / det;
   // Calculate inverse using adjoint matrices
   OVm3 m;
   m[ 0] =  Det2x2(b2, b3, c2, c3) * invdet;
   m[ 1] = -Det2x2(a2, a3, c2, c3) * invdet;
   m[ 2] =  Det2x2(a2, a3, b2, b3) * invdet;
   m[ 3] = -Det2x2(b1, b3, c1, c3) * invdet;
   m[ 4] =  Det2x2(a1, a3, c1, c3) * invdet;
   m[ 5] = -Det2x2(a1, a3, b1, b3) * invdet;
   m[ 6] =  Det2x2(b1, b2, c1, c2) * invdet;
   m[ 7] = -Det2x2(a1, a2, c1, c2) * invdet;
   m[ 8] =  Det2x2(a1, a2, b1, b2) * invdet;
   return m;
} 



//=============================================================//
//                                                             //
//                       4x4 Matrix                            //
//                                                             //
//=============================================================//

/*! \brief Constructor from 16 elements.
 *  \param eXX = eRC, element (R,C) in matrix. */
OVm4::OVm4(float e00,float e01,float e02,float e03,
           float e10,float e11,float e12,float e13,
           float e20,float e21,float e22,float e23,
           float e30,float e31,float e32,float e33) {
	float a[16] = {e00,e01,e02,e03,
	               e10,e11,e12,e13,
	               e20,e21,e22,e23,
	               e30,e31,e32,e33};
	this->Set(a);
}

/*! \brief Sets matrix elements.
 *  \param a Array of 16 floats. */
void OVm4::Set(float *a) {
	*this = a;
}

/*! \brief Access operator for entire row of matrix.
 *  \param index Row index.
 *  \return Returns indexed row as OV3f vector. */
OVv4 OVm4::GetRow(int index) {
    index = (index < 0) ? 0 : ((index > 3) ? 3 : index);
	float row[4] = {e[4*index+0], e[4*index+1], e[4*index+2], e[4*index+3]};
	return OVv4(row);
}

/*! \brief Access operator for entire column of matrix.
 *  \param index Column index.
 *  \return Returns indexed column as OV3f vector. */
OVv4 OVm4::GetCol(int index) {
    index = (index < 0) ? 0 : ((index > 3) ? 3 : index);
	float col[4] = {e[0+index], e[4+index], e[8+index], e[12+index]};
	return OVv4(col);
}

/*! \brief Sets one row of matrix elements.
 *  \param index Row index. 
 *  \param v Contains the elements to which the row should be set
 */
void OVm4::SetRow(int index,float *v) {
    index = (index < 0) ? 0 : ((index > 3) ? 3 : index);
	e[4*index  ] = v[0];
	e[4*index+1] = v[1];
	e[4*index+2] = v[2];
	e[4*index+3] = v[3];
}

/*! \brief Sets one column of matrix elements.
 *  \param index Column index. 
 *  \param v Contains the elements to which the column should be set
 */
void OVm4::SetCol(int index,float *v) {
    index = (index < 0) ? 0 : ((index > 3) ? 3 : index);
	e[   index] = v[0];
	e[4 +index] = v[1];
	e[8 +index] = v[2];
	e[12+index] = v[3];
}

/*! \brief Clears matrix to 0. */
void OVm4::Clear() {
	for(int i=0;i<16;i++)
		e[i] = 0;
}

/*! \brief Sets matrix to identity. */
void OVm4::SetToIdentityMatrix() {
	Clear();
	e[ 0] = 1;
	e[ 5] = 1;
	e[10] = 1;
	e[15] = 1;
}

/*! \brief Sets matrix to translation matrix.
 *  Multiplication will perform a translation of (tx,ty,tz) units.
 *  \param tx Translation in x.
 *  \param ty Translation in y.
 *  \param tz Translation in z.
 */
void OVm4::SetToTranslationMatrix(float tx,float ty,float tz) {
	Clear();
	e[0] = tx;
	e[5] = ty;
	e[10] = tz;
	e[15] = 1;
}

/*! \brief Sets matrix to rotation matrix.
 *  Multiplication will perform a rotation of theta degrees (radians?) around (x,y,z) axis.
 *  \param theta Angle of rotation.
 *  \param x x-coordinate of rotation axis.
 *  \param y y-coordinate of rotation axis.
 *  \param z y-coordinate of rotation axis.
 */
void OVm4::SetToRotationMatrix(float theta, float x, float y, float z) {
	Clear();
	float c = cos(theta);
	float s = sin(theta);
	float t = 1 - cos(theta);     
	e[0] = t*x*x + c;
	e[1] = t*x*y + s*z;
	e[2] = t*x*z - s*y;
	e[4] = t*x*y - s*z;
	e[5] = t*y*y + c;
	e[6] = t*y*z + s*x;
	e[8] = t*x*z + s*y;
	e[9] = t*y*z - s*x;
	e[10] = t*z*z + c;
	e[15] = 1;
}

/*! \brief Prints matrix to commandline. */
void OVm4::Print() {
	std::printf("%f %f %f %f\n%f %f %f %f\n%f %f %f %f\n%f %f %f %f\n\n",
		e[0],e[1],e[2],e[3],e[4],e[5],e[6],e[7],e[8],e[9],e[10],e[11],e[12],e[13],e[14],e[15]);
}

/*! \brief << standard C ostream operator.
 *  \param out Reference to active ostream.
 *  \param m Reference to matrix.
 *  \return Returns the ostream reference in parameter out. */
std::ostream& operator<<(std::ostream &out, const OVm4 &m)
{
	out << m[ 0] << " " << m[ 1] << " " << m[ 2] << " " << m[ 3] << std::endl;
	out << m[ 4] << " " << m[ 5] << " " << m[ 6] << " " << m[ 7] << std::endl;
	out << m[ 8] << " " << m[ 9] << " " << m[10] << " " << m[11] << std::endl;
	out << m[12] << " " << m[13] << " " << m[14] << " " << m[15];
	return out;
}

/*! \brief Operator for multiplication with scalar.
 *  \param f Constant scalar.
 *  \return Returns result of multiplication as OVm4 matrix. */
OVm4 OVm4::operator*(const float f) {
	OVm4 m;
	for(int i=0;i<16;i++)
		m[i] = e[i]*f;
	return m;
}

/*! \brief Operator for multiplication with vector.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OV4f.
 *  \param v Constant reference to vector.
 *  \return Returns result of multiplication as OV4f vector. */
OVv4 OVm4::operator*(const OVv4 &v) {
	OVv4 vnew;
	vnew[0] = e[ 0]*v[0] + e[ 1]*v[1] + e[ 2]*v[2] + e[ 3]*v[3];
	vnew[1] = e[ 4]*v[0] + e[ 5]*v[1] + e[ 6]*v[2] + e[ 7]*v[3];
	vnew[2] = e[ 8]*v[0] + e[ 9]*v[1] + e[10]*v[2] + e[11]*v[3];
	vnew[3] = e[12]*v[0] + e[13]*v[1] + e[14]*v[2] + e[15]*v[3];
	return vnew;
}

/*! \brief Operator for multiplication with 4x4 matrix.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OVm4.
 *  \param m Constant reference to matrix.
 *  \return Returns result of multiplication as OVm4 matrix. */
OVm4 OVm4::operator*(const OVm4 &a) {
	OVm4 m;     
	m[ 0] = e[ 0]*a[0] + e[ 1]*a[4] + e[ 2]*a[ 8] + e[ 3]*a[12];
	m[ 1] = e[ 0]*a[1] + e[ 1]*a[5] + e[ 2]*a[ 9] + e[ 3]*a[13];
	m[ 2] = e[ 0]*a[2] + e[ 1]*a[6] + e[ 2]*a[10] + e[ 3]*a[14];
	m[ 3] = e[ 0]*a[3] + e[ 1]*a[7] + e[ 2]*a[11] + e[ 3]*a[15];
	m[ 4] = e[ 4]*a[0] + e[ 5]*a[4] + e[ 6]*a[ 8] + e[ 7]*a[12];
	m[ 5] = e[ 4]*a[1] + e[ 5]*a[5] + e[ 6]*a[ 9] + e[ 7]*a[13];
	m[ 6] = e[ 4]*a[2] + e[ 5]*a[6] + e[ 6]*a[10] + e[ 7]*a[14];
	m[ 7] = e[ 4]*a[3] + e[ 5]*a[7] + e[ 6]*a[11] + e[ 7]*a[15];
	m[ 8] = e[ 8]*a[0] + e[ 9]*a[4] + e[10]*a[ 8] + e[11]*a[12];
	m[ 9] = e[ 8]*a[1] + e[ 9]*a[5] + e[10]*a[ 9] + e[11]*a[13];
	m[10] = e[ 8]*a[2] + e[ 9]*a[6] + e[10]*a[10] + e[11]*a[14];
	m[11] = e[ 8]*a[3] + e[ 9]*a[7] + e[10]*a[11] + e[11]*a[15];
	m[12] = e[12]*a[0] + e[13]*a[4] + e[14]*a[ 8] + e[15]*a[12];
	m[13] = e[12]*a[1] + e[13]*a[5] + e[14]*a[ 9] + e[15]*a[13];
	m[14] = e[12]*a[2] + e[13]*a[6] + e[14]*a[10] + e[15]*a[14];
	m[15] = e[12]*a[3] + e[13]*a[7] + e[14]*a[11] + e[15]*a[15];     
	return m;
}

/*! \brief = operator.
 *  This function could easily be re-implemented using an array of floats
 *  instead of OVm3.
 *  \param m Constant reference to matrix. */
void OVm4::operator=(const float *a) {
	for(int i=0;i<16;i++)
		e[i] = a[i];
}

/*! \brief Determinant of any 3x3 matrix.
 *  \param eXX eXX = eRC, element (R,C) in matrix.
 *  \return Scalar value of determinant. */
float OVm4::Det3x3(float &e00,float &e01,float &e02,
		                  float &e10,float &e11,float &e12,
                          float &e20,float &e21,float &e22) {
	return (e00*(e11*e22 - e12*e21) +
			e01*(e12*e20 - e10*e22) +
			e02*(e10*e21 - e11*e20));
}

/*! \brief Determinant of any 4x4 matrix.
 *  Using piecewise 3x3 determinants.
 *  \param eXX eXX = eRC, element (R,C) in matrix.
 *  \return Scalar value of determinant. */
float OVm4::Det4x4(float &e00,float &e01,float &e02,float &e03,
		                  float &e10,float &e11,float &e12,float &e13,
                          float &e20,float &e21,float &e22,float &e23,
			              float &e30,float &e31,float &e32,float &e33) {
	return ( e00*Det3x3(e11,e12,e13, e21,e22,e23, e31,e32,e33)
			-e01*Det3x3(e10,e12,e13, e20,e22,e23, e30,e32,e33)
			+e02*Det3x3(e10,e11,e13, e20,e21,e23, e30,e31,e33)
			-e03*Det3x3(e10,e11,e12, e20,e21,e22, e30,e31,e32) );
}

/*! \brief Determinant of object matrix.
 *  \return Scalar value of determinant. */
float OVm4::GetDeterminant() {
	return Det4x4(e[ 0],e[ 1],e[ 2],e[ 3],
				  e[ 4],e[ 5],e[ 6],e[ 7],
				  e[ 8],e[ 9],e[10],e[11],
				  e[12],e[13],e[14],e[15]);
}

/*! \brief Transpose of object matrix.
 *  \return Transpose OVm4 matrix. */
OVm4 OVm4::GetTranspose() {
	OVm4 m;
	m[ 0] = e[ 0];
	m[ 1] = e[ 4];
	m[ 2] = e[ 8];
	m[ 3] = e[12];
	m[ 4] = e[ 1];
	m[ 5] = e[ 5];
	m[ 6] = e[ 9];
	m[ 7] = e[13];
	m[ 8] = e[ 2];
	m[ 9] = e[ 6];
	m[10] = e[10];
	m[11] = e[14];
	m[12] = e[ 3];
	m[13] = e[ 7];
	m[14] = e[11];
	m[15] = e[15];
	return m;
}

/*! \brief Inverse of object matrix.
 *  Using adjoint transpose matrices to calculate inverse.
 *  \return Inverse OVm4 matrix. */
OVm4 OVm4::GetInverse() {
	// Set up transpose
   float a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;
   a1 = e[0]; b1 = e[4]; c1 = e[ 8]; d1 = e[12];
   a2 = e[1]; b2 = e[5]; c2 = e[ 9]; d2 = e[13];
   a3 = e[2]; b3 = e[6]; c3 = e[10]; d3 = e[14];
   a4 = e[3]; b4 = e[7]; c4 = e[11]; d4 = e[15];
   float det = Det4x4(a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4);
   // Check if singular matrix
   if(fabs(det) < 0.0000001) {
   		printf("(OVm4.inverse): Singular matrix, unable to calculate inverse\n");
   		return OVm4();
   }
   float invdet = 1.0f / det;
   // Calculate inverse using adjoint matrices
   OVm4 m;
   m[ 0] =  Det3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4) * invdet;
   m[ 1] = -Det3x3(a2, a3, a4, c2, c3, c4, d2, d3, d4) * invdet;
   m[ 2] =  Det3x3(a2, a3, a4, b2, b3, b4, d2, d3, d4) * invdet;
   m[ 3] = -Det3x3(a2, a3, a4, b2, b3, b4, c2, c3, c4) * invdet;
   m[ 4] = -Det3x3(b1, b3, b4, c1, c3, c4, d1, d3, d4) * invdet;
   m[ 5] =  Det3x3(a1, a3, a4, c1, c3, c4, d1, d3, d4) * invdet;
   m[ 6] = -Det3x3(a1, a3, a4, b1, b3, b4, d1, d3, d4) * invdet;
   m[ 7] =  Det3x3(a1, a3, a4, b1, b3, b4, c1, c3, c4) * invdet;
   m[ 8] =  Det3x3(b1, b2, b4, c1, c2, c4, d1, d2, d4) * invdet;
   m[ 9] = -Det3x3(a1, a2, a4, c1, c2, c4, d1, d2, d4) * invdet;
   m[10] =  Det3x3(a1, a2, a4, b1, b2, b4, d1, d2, d4) * invdet;
   m[11] = -Det3x3(a1, a2, a4, b1, b2, b4, c1, c2, c4) * invdet;
   m[12] = -Det3x3(b1, b2, b3, c1, c2, c3, d1, d2, d3) * invdet;
   m[13] =  Det3x3(a1, a2, a3, c1, c2, c3, d1, d2, d3) * invdet;
   m[14] = -Det3x3(a1, a2, a3, b1, b2, b3, d1, d2, d3) * invdet;
   m[15] =  Det3x3(a1, a2, a3, b1, b2, b3, c1, c2, c3) * invdet;
   return m;
} 

}  // End of OV
