/*! \file ovalgebra.h
 *  \brief Header file for OVvX and OVmX.
 *
 *  Contains the declarations of vector and matrix containers. 
 */

#ifndef _OVALGEBRA_H_
#define _OVALGEBRA_H_

#include <ostream>
#include <math.h>
#ifndef uint
	typedef unsigned int uint;
#endif

namespace OV
{



/*! \brief Template for float vectors.
 *  Container for X floats with access operators
 */
template<int X>
struct OVvX
{
	float e[X]; /*!< Array of elements. */
	/*! \brief Empty constructor. */
	OVvX() {}
    /*! \brief Constructor from array of 2 floats. */
    OVvX(float *a) {*this = a;}
	/*! \brief Destructor. */
	~OVvX() {}
    /*! \brief Automatic typecast operator.
     *  Enables use of OV2f in glFunc[v]() functions.
     *  \return Returns pointer to first element in array */
	inline operator float*() {return e;}
	/*! \brief Access operator.
     *  \param index Index of element.
     *  \return Returns indexed element */
	inline float& operator[](int index) {return e[index];}
    /*! \brief Constant access operator.
     *  \param index Index of element.
     *  \return Returns indexed element */
	inline const float& operator[](int index) const {return e[index];}
	/*! \brief = operator.
	 *  \param a Reference to other object. */
	inline void operator=(const float *a) {for(int i=0;i<X;i++) e[i]=a[i];}
};

/*! \brief Typedef for OV2f. */
typedef OVvX<2> OVv2;
/*! \brief Typedef for OV3f. */
typedef OVvX<3> OVv3;
/*! \brief Typedef for OV4f. */
typedef OVvX<4> OVv4;
/*! \brief Typedef for OV9f. */
typedef OVvX<9> OVv9;
/*! \brief Typedef for OV16f. */
typedef OVvX<16> OVv16;



/*! \brief 3x3 floating point matrix class.
 *  Container for 3x3 floating point matrices with access operators.
 */
class OVm3
{
	public:
        /*! \brief Empty constructor, sets identity matrix */
		OVm3() {this->SetToIdentityMatrix();}
    	/*! \brief Constructor from array of 9 floats. */	
		OVm3(float *m) {*this = m;}
		OVm3(float e00,float e01,float e02,
			 float e10,float e11,float e12,
             float e20,float e21,float e22);
        /*! \brief Destructor. */
		~OVm3() {}
        /*! \brief Automatic typecast operator.
         *  Enables use of OVm3 in glFunc[v]() functions.
         *  \return Returns pointer to first element in matrix */
		operator float*() {return e;}
    	/*! \brief Linear access operator.
		 *  'int' and 'unsigned int' as well as const declared versions avaliable.
         *  \param index Linear index of element.
         *  \return Returns indexed element. */
		float& operator[](int index) {return e[index];}
		float& operator[](unsigned int index) {return e[index];}
		const float& operator[](int index) const {return e[index];}
		const float& operator[](unsigned int index) const {return e[index];}
    	/*! \brief Access operator.
		 *  'int' and 'unsigned int' as well as const declared versions avaliable.
         *  \param row Row index of element.
         *  \param col Column index of element.
         *  \return Returns indexed element. */
		float& operator()(int row,int col) {return e[3*row+col];}
		float& operator()(unsigned int row,unsigned int col) {return e[3*row+col];}
		const float& operator()(int row,int col) const {return e[3*row+col];}
		const float& operator()(unsigned int row,unsigned int col) const {return e[3*row+col];}

		OVm3 operator*(const float f);
		OVv3 operator*(const OVv3 &v);
		OVm3 operator*(const OVm3 &a);
		void operator=(const float *a);

    	/*! \brief Use only if OV9f is wanted, else use overloaded typecast (OVm3 -> float*).
			\return Returns elements as vector (compatible with float*). */
		OVv9 Get() {return OVv9(e);};
		OVv3 GetRow(int index);
		OVv3 GetCol(int index);
		void Set(float *a);
		void SetRow(int index,float *v);
		void SetCol(int index,float *v);

		void Print();
		void Clear();
		void SetToIdentityMatrix();
		
		float GetDeterminant();
		OVm3 GetTranspose();
		OVm3 GetInverse();
	protected:
		inline float Det2x2(float &e00,float &e01,
                            float &e10,float &e11);
		inline float Det3x3(float &e00,float &e01,float &e02,
					        float &e10,float &e11,float &e12,
                            float &e20,float &e21,float &e22);
	private:
		float e[9]; /*!< Array of matrix elements. */
};



/*! \brief 4x4 floting point matrix class.
 *  Container for 4x4 floating point matrices with access operators.
 */
class OVm4
{
	public:
        /*! \brief Empty constructor, sets identity matrix */	
		OVm4() {this->SetToIdentityMatrix();}
    	/*! \brief Constructor from array of 16 floats. */			
		OVm4(float *m) {*this = m;}
		OVm4(float e00,float e01,float e02,float e03,
		     float e10,float e11,float e12,float e13,
             float e20,float e21,float e22,float e23,
             float e30,float e31,float e32,float e33);
        /*! \brief Destructor. */ 
		~OVm4() {}
        /*! \brief Automatic typecast operator.
         *  Enables use of OVm4 in glFunc[v]() functions.
         *  \return Returns pointer to first element in matrix */		
		operator float*() {return e;}
    	/*! \brief Linear access operator.
		 *  'int' and 'unsigned int' as well as const declared versions avaliable.
         *  \param index Linear index of element.
         *  \return Returns indexed element. */		
		float& operator[](int index) {return e[index];}
		float& operator[](unsigned int index) {return e[index];}
		const float& operator[](int index) const {return e[index];}
		const float& operator[](unsigned int index) const {return e[index];}
    	/*! \brief Access operator.
		 *  'int' and 'unsigned int' as well as const declared versions avaliable.
         *  \param row Row index of element.
         *  \param col Column index of element.
         *  \return Returns indexed element. */		
		float& operator()(int row,int col) {return e[4*row+col];}
		float& operator()(unsigned int row,unsigned int col) {return e[4*row+col];}
		const float& operator()(int row,int col) const {return e[4*row+col];}
		const float& operator()(unsigned int row,unsigned int col) const {return e[4*row+col];}
		
		OVm4 operator*(const float f);
		OVv4 operator*(const OVv4 &v);
		OVm4 operator*(const OVm4 &a);
		void operator=(const float *a);
		
    	/*! \brief Use only if OV16f is wanted, else use overloaded typecast (OVm4 -> float*).
			\return Returns elements as vector (compatible with float*). */
		OVv16 Get() {return OVv16(e);}
		OVv4 GetRow(int index);
		OVv4 GetCol(int index);
		void Set(float *a);
		void SetRow(int index,float *v);
		void SetCol(int index,float *v);

		void Print();
		void Clear();
		void SetToIdentityMatrix();
		void SetToTranslationMatrix(float tx,float ty,float tz);
		void SetToRotationMatrix(float theta,float x,float y,float z);

		float GetDeterminant();
		OVm4 GetTranspose();
		OVm4 GetInverse();
	protected:
		inline float Det3x3(float &e00,float &e01,float &e02,
					        float &e10,float &e11,float &e12,
                            float &e20,float &e21,float &e22);
        inline float Det4x4(float &e00,float &e01,float &e02,float &e03,
				            float &e10,float &e11,float &e12,float &e13,
                            float &e20,float &e21,float &e22,float &e23,
                            float &e30,float &e31,float &e32,float &e33);
	private:
		float e[16]; /*!< Array of matrix elements. */
};

}  // End of OV

#endif

