/*! \file ovmesh.cpp
 *  \brief OVMesh source file.
 *
 */

#include "ovmesh.h"

namespace OV
{

#define odd(x) (x%2)

/*! \brief Constructor for ovmesh.
 *
 */
OVMesh::OVMesh() 
{
    m_numVertices = 0;
    m_numFaces = 0;
}

/*! \brief Destructor for ovmesh.
 *
 */
OVMesh::~OVMesh()
{
    m_Clear();
}

/*! \brief Generates a grid.
 *
 *  Generates a rectangular grid between (x0,y0) and (x1,y1). If vertical is
 *  TRUE the grid will be in the xy-plane, otherwise it will be in the
 *  xz-plane.
 *
 *  \param resX the resolution in the x direction.
 *  \param resY the resolution in the y(z) direction.
 *  \param x0 the x component of the starting point.
 *  \param x1 the x component of the ending point.
 *  \param y0 the y component of the starting point.
 *  \param y1 the y component of the ending point.
 *  \param vertical decides if the grid is vertical or horizontal.
 *  \sa m_AddVertex()
 */
void OVMesh::GenerateGrid(int resX, int resY, float x0, float x1, float y0,
						float y1, bool vertical)
{
    // Clear already loaded ovmesh
    m_Clear();

    // Generate vertex list
    float dx = (x1-x0) / (resX-1);
    float dy = (y1-y0) / (resY-1);

    m_vertArray = new GLfloat[resX*resY*3];
    m_numVertices = 0;

    for(int y=0; y<resY; y++)
    {
        for(int x=0; x<resX; x++)
        {
            if(vertical)
                m_AddVertex(x0+x*dx, y0+y*dy, 0.0f);
            else
                m_AddVertex(x0+x*dx, 0.0f, y0+y*dy);
        }
    }

    // Generate index list
    m_faceArray = new GLuint[2*(resY-1)*resX];
    
    int i=0;

    for(int y=0; y<resY-1; y++)
    {
        if(odd(y))
        {
            for(int x=resX-1; x>=0; x--)
            {
                m_faceArray[i++] = (resX*(y+1)+x);
                m_faceArray[i++] = (resX*y+x);
            }
        }
        else
        {
            for(int x=0; x<resX; x++)
            {
                m_faceArray[i++] = (resX*y+x);
                m_faceArray[i++] = (resX*(y+1)+x);
            }
        }
    }
    m_numFaces = i-2;
}

/*! \brief Creates a Vertex Buffer Object (VBO)
 *
 */
void OVMesh::CreateVBO()
{
    // Vertices
    glGenBuffersARB(1, &m_vertexVBO);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, m_numVertices*3*sizeof(GLfloat),
                        m_vertArray, GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

    // Indices
    glGenBuffersARB(1, &m_indexVBO);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexVBO);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, (m_numFaces+2)*
                    sizeof(GLuint), m_faceArray, GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
}

/*! \brief Deletes a Vertex Buffer Object (VBO)
 *
 */
void OVMesh::DeleteVBO()
{
    glDeleteBuffersARB(1, &m_vertexVBO);
    glDeleteBuffersARB(1, &m_indexVBO);
}

/*! \brief Draws the mesh using OpenGL. 
 *
 *  Draws the mesh using glDrawElements and offline arrays of vertices.
 */
void OVMesh::DrawOpenGL()
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, m_vertArray);
    glDrawElements(GL_TRIANGLE_STRIP, m_numFaces+2, GL_UNSIGNED_INT,
                   m_faceArray);
    glDisableClientState(GL_VERTEX_ARRAY);
}

/*! \brief Draws the mesh as a Vertex Buffer Object (VBO)
 *
 *  Draws the mesh using glDrawElements and a VBO stored in video RAM.
 */
void OVMesh::DrawVBO()
{
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVBO);
    glVertexPointer(3, GL_FLOAT, 0, 0);
    glEnableClientState(GL_VERTEX_ARRAY);

    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexVBO);

    glDrawElements(GL_TRIANGLE_STRIP, m_numFaces+2, GL_UNSIGNED_INT, 0);

    glDisableClientState(GL_VERTEX_ARRAY);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
}

/*! \brief Calculates the number of triangles (faces) in the grid.
 *
 *  \return The number of triangles.
 */
unsigned int OVMesh::GetNumTriangles()
{
    return m_numFaces;
}

/*! \brief Adds a vertex in the list of vertices.
 *
 */
void OVMesh::m_AddVertex(float x, float y, float z)
{
    int index = m_numVertices*3;
    m_vertArray[index] = x;
    m_vertArray[index+1] = y;
    m_vertArray[index+2] = z;
    m_numVertices++;
}

/*! \brief Clears the vertex list and the face list.
 *
 */
void OVMesh::m_Clear()
{
    if(m_vertArray)
        delete[] m_vertArray;
    if(m_faceArray)
        delete[] m_faceArray;
}

} // End of OV
