/*! \file ovprojector.cpp
 *  \brief OVProjector source file.
 *
 *  Contains the definitions of the OVProjector class. Calculates the 
 *  MRange matrix by setting the projection frustum from the camera
 *  primitives. Also defines and the projector's modelview and projection
 *  matrices which are used for projection of the mesh defined in image space. 
 *  There are also methods for setting the current modelview matrix and/or the
 *  projection matrix according to these matrices.
 */

#include "ovprojector.h"

namespace OV
{

/*! \brief The constructor for the OVProjector class. */
OVProjector::OVProjector()
{    
    m_projPos = new float[3];
    m_projDir = new float[3];
    m_projUp = new float[3];
}

/*! \brief The destructor of the OVProjector class. */
OVProjector::~OVProjector()
{    
    delete[] m_projPos;
    delete[] m_projDir;
    delete[] m_projUp;
}

/*! \brief Access function for range matrix.
 *
 *  \return Pointer to array of 16 floats (range matrix). 
 */
float* OVProjector::GetMRange()
{
    return m_mRange;
}

/*! \brief Calculates projection frustum from camera primitives. */
void OVProjector::m_CalcProjectionFrustum()
{
    //The projector's up vector is always (0,1,0)
    m_projUp[0] = 0.0;
    m_projUp[1] = 1.0;
    m_projUp[2] = 0.0;
    
    //Position the projector
    m_projPos[0] = m_camPos[0] - 5.0f*m_camDir[0]; //x coordinate from the camera
    m_projPos[2] = m_camPos[2] - 5.0f*m_camDir[2]; //y coordinate from the camera

    //Place the projector at least 10 m. above the upper displacement plane
    m_projPos[1] = m_camPos[1] - 5.0f*m_camDir[1];

    m_projDir[0] = m_camDir[0]; //x coordinate from the camera
    m_projDir[1] = m_camDir[1]; //y coordinate from the camera
    m_projDir[2] = m_camDir[2]; //z coordinate from the camera
}

/*! \brief Calculates intersections between projector frustum and water volume.
 *
 *  Result is stored in member m_intersections.
 */
void OVProjector::m_CalcFrustumWaterIntersections()
{
    //Get the current modelview matrix (the projector's)
    GLdouble modelView[16]; 
    glGetDoublev(GL_MODELVIEW_MATRIX, modelView);
    
    //Get the current projection matrix (the projector's)
    GLdouble projection[16];
    glGetDoublev(GL_PROJECTION_MATRIX, projection);  

    //Get the current viewport matrix
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    
    //Transform the frustum cornerpoints on the far clip plane to world coords
    OVv4 p1WorldSpace;
    GLdouble px=0.0, py=0.0, pz=0.0;

    gluUnProject(0.0, 0.0, 1.0, modelView, projection, viewport, &px, &py, &pz);
    p1WorldSpace[0] = px;
    p1WorldSpace[1] = py;
    p1WorldSpace[2] = pz;
    p1WorldSpace[3] = 1.0;

    OVv4 p2WorldSpace;  
    gluUnProject(0.0, (GLdouble)(*m_windowHeight), 1.0, modelView, projection, 
    viewport, &px, &py, &pz);
    p2WorldSpace[0] = px;
    p2WorldSpace[1] = py;
    p2WorldSpace[2] = pz;
    p2WorldSpace[3] = 1.0;

    OVv4 p3WorldSpace;  
    gluUnProject((GLdouble)(*m_windowWidth), (GLdouble)(*m_windowHeight), 1.0, 
    modelView, projection, viewport, &px, &py, &pz);
    p3WorldSpace[0] = px;
    p3WorldSpace[1] = py;
    p3WorldSpace[2] = pz;
    p3WorldSpace[3] = 1.0;

    OVv4 p4WorldSpace;
    gluUnProject((GLdouble)(*m_windowWidth), 0.0, 1.0, modelView, projection, 
    viewport,  &px, &py, &pz);    
    p4WorldSpace[0] = px;
    p4WorldSpace[1] = py;
    p4WorldSpace[2] = pz;
    p4WorldSpace[3] = 1.0;

    std::vector<OVv4> frustumCornerPointsFront;

    frustumCornerPointsFront.push_back(p1WorldSpace);
    frustumCornerPointsFront.push_back(p2WorldSpace);
    frustumCornerPointsFront.push_back(p3WorldSpace);
    frustumCornerPointsFront.push_back(p4WorldSpace);

    //Calculate the intersections between the line from the camera to the 
    //cornerpoint and the upper and lower displacement planes
    //and calculate the intersections between the lines between the cornerpoints 
    //and the upper and lower displacement planes
    //Then project the intersections onto the base plane    

    OVv4 projPos;
    projPos[0] = m_projPos[0];
    projPos[1] = m_projPos[1];
    projPos[2] = m_projPos[2];
    projPos[3] = 1.0f;

    OVv4 projDir;
    projDir[0] = m_projDir[0];
    projDir[1] = m_projDir[1];
    projDir[2] = m_projDir[2];
    projDir[3] = 1.0f;

    //Remove the old intersections
    m_intersectionsWorldSpace.clear();

    //Intersections for the front cornerpoints (i.e. on the far plane, 
    //ignore the near plane)
    for(unsigned int i = 0; i < frustumCornerPointsFront.size(); i++)
    {
        //Calculate the intersections between the line from the camera to the 
        //cornerpoint and the upper and lower displacement planes       

        //Calculate intersection with the upper displacement plane
        OVv4 intersectionUpper; 
        bool intersectUpper = m_CalcWaterplaneYLineIntersection(
            (*m_maxDisplacement) + m_baseplaneY, projPos,
            frustumCornerPointsFront[i], &intersectionUpper);

        if(intersectUpper == true)
        {
            //Project onto the baseplane
            intersectionUpper[1] = m_baseplaneY;
            //Add intersection to m_intersectionsWorldSpace
            m_intersectionsWorldSpace.push_back(intersectionUpper);
        }

        //Calculate intersection with the lower displacement plane
        OVv4 intersectionLower; 
        bool intersectLower = m_CalcWaterplaneYLineIntersection(
            -(*m_maxDisplacement) + m_baseplaneY, projPos,
            frustumCornerPointsFront[i], &intersectionLower);

        if(intersectLower == true)
        {
            //Project onto the baseplane
            intersectionLower[1] = m_baseplaneY;
            //Add intersection to m_intersectionsWorldSpace
            m_intersectionsWorldSpace.push_back(intersectionLower);
        }

        //Calculate the intersections between the lines between the cornerpoints
        // and the upper and lower displacement planes
        
        //Calculate intersection with the upper displacement plane
        OVv4 intersectionUpperSide; 
        bool intersectUpperSide;
        if(i == frustumCornerPointsFront.size()-1)
        {             
            intersectUpperSide = m_CalcWaterplaneYLineIntersection(
                (*m_maxDisplacement)+m_baseplaneY, frustumCornerPointsFront[i], 
                frustumCornerPointsFront[0], &intersectionUpperSide);
        }
        else
        {            
            intersectUpperSide = m_CalcWaterplaneYLineIntersection(
                (*m_maxDisplacement)+m_baseplaneY, frustumCornerPointsFront[i], 
                frustumCornerPointsFront[i+1], &intersectionUpperSide);
        }
        if(intersectUpperSide == true)
        {
            //Project onto the baseplane
            intersectionUpperSide[1] = m_baseplaneY;
            //Add intersection to m_intersectionsWorldSpace
            m_intersectionsWorldSpace.push_back(intersectionUpperSide);
        }
        
        //Calculate intersection with the lower displacement plane
        OVv4 intersectionLowerSide; 
        bool intersectLowerSide;
        if(i == frustumCornerPointsFront.size()-1)
        {             
            intersectLowerSide = m_CalcWaterplaneYLineIntersection(
                -(*m_maxDisplacement)+m_baseplaneY, frustumCornerPointsFront[i],
                frustumCornerPointsFront[0], &intersectionLowerSide);
        }
        else
        {            
            intersectLowerSide = m_CalcWaterplaneYLineIntersection(
                -(*m_maxDisplacement)+m_baseplaneY, frustumCornerPointsFront[i],
                frustumCornerPointsFront[i+1], &intersectionLowerSide);
        }
        if(intersectLowerSide == true)
        {
            //Project onto the baseplane
            intersectionLowerSide[1] = m_baseplaneY;
            //Add intersection to m_intersectionsWorldSpace
            m_intersectionsWorldSpace.push_back(intersectionLowerSide);
        }

        //If the cornerpoint lies between the upper and lower displacement plane,
        //it is also of interest for the range matrix, and should be stored
        float y = (frustumCornerPointsFront[i])[1];
        if((y >(-(*m_maxDisplacement) + m_baseplaneY))
            && (y < ((*m_maxDisplacement) + m_baseplaneY)))
        {
            OVv4 corner(frustumCornerPointsFront[i]);
            //Project onto baseplane
            corner[1] = m_baseplaneY;
            m_intersectionsWorldSpace.push_back(corner);
        }
    }
}

/*! \brief Calculates the intersection between a line and a plane 
*          parallell to the xz plane.
*
*  Calculates the intersection between a line (point1->point2) 
*  and a plane (in the xz plane) with y-coordinate planeY.
*  The resulting intersection, if any, is stored in intersection.
*
*  \param planeY The y-coordinate of the plane
*  \param point1 The point where the line begins
*  \param point2 The point where the line ends
*
*  \return True if an intersection is found.
*/
bool OVProjector::m_CalcWaterplaneYLineIntersection(float planeY, OVv4 point1, 
    OVv4 point2, OVv4* intersection)
{
    if(point2[1] != point1[1])
    {
        float k = (planeY - point1[1]) / (point2[1] - point1[1]);
    
        float minY = point1[1];
        float maxY = point2[1];
        if(maxY < minY)
        {
            float temp = minY;
            minY = maxY;
            maxY = temp;
        }
        
        if((k >= 0) && (minY <= planeY) && (maxY >= planeY))
        {
            (*intersection)[0] = point1[0] + k*(point2[0] - point1[0]);

            //Project onto the plane
            (*intersection)[1] = planeY;
            (*intersection)[2] = point1[2] + k*(point2[2] - point1[2]);
            (*intersection)[3] = 1.0f;

            return true;
        } 
    }
    return false; 
}

/*! \brief Calculates min max values used by MRange matrix.
 *
 *  Projects world space intersections to image space and calculates min/max
 *  values for x and y coordinates.
 *  Resaults are stored in m_minImSpaceX etc.
 */
void OVProjector::m_CalcImSpaceMinMax()
{
    //Get the current perspective matrix (the projector's)
    GLdouble persp[16]; 
    glGetDoublev(GL_PROJECTION_MATRIX, persp);

    //Get the current modelview matrix (the projector's)
    GLdouble modelView[16]; 
    glGetDoublev(GL_MODELVIEW_MATRIX, modelView);        

    //Get the current viewport matrix
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    
    //Transform the intersection points from world space to image space 
    //(defined by the projector)

    //clear the image space vector
    m_intersectionsImageSpace.clear();     

    std::vector<OVv4>::iterator it;
    for( it = m_intersectionsWorldSpace.begin(); 
         it != m_intersectionsWorldSpace.end(); it++)
    {
        GLdouble winx2 =0.0;
        GLdouble winy2 =0.0;
        GLdouble winz2 =0.0;
        gluProject((*it)[0], (*it)[1], (*it)[2], modelView, persp, viewport, 
        &winx2, &winy2, &winz2);
        
        OVv4 intersectionImageSpace2;
        intersectionImageSpace2[0] = (2.0*(winx2-viewport[0])/viewport[2])-1.0;
        intersectionImageSpace2[1] = (2.0*(winy2-viewport[1])/viewport[3])-1.0;
        intersectionImageSpace2[2] = 2.0*winz2-1.0;
        intersectionImageSpace2[3] = 1.0;
        m_intersectionsImageSpace.push_back(intersectionImageSpace2);
    }     
    
    if(m_intersectionsImageSpace.size() > 0)
    {       
        std::vector<OVv4>::iterator it = m_intersectionsImageSpace.begin();
        m_minImSpaceX = 1.0; 
        m_minImSpaceY = 1.0;  
        m_maxImSpaceX = -1.0;  
        m_maxImSpaceY = -1.0;

        for(;it != m_intersectionsImageSpace.end(); it++)
        {            
            if((*it)[0] < m_minImSpaceX) m_minImSpaceX = (*it)[0];
            if((*it)[1] < m_minImSpaceY) m_minImSpaceY = (*it)[1];
            if((*it)[0] > m_maxImSpaceX) m_maxImSpaceX = (*it)[0];
            if((*it)[1] > m_maxImSpaceY) m_maxImSpaceY = (*it)[1];
        }        
        
        m_minImSpaceX -= 0.01;
        m_minImSpaceY -= 0.01;
        m_maxImSpaceX += 0.01;
        m_maxImSpaceY += 0.01;
    }
    else
    {
        m_minImSpaceX = -1.0; 
        m_minImSpaceY = -1.0;   
        m_maxImSpaceX = 1.0;   
        m_maxImSpaceY = 1.0;    
    }

    //Clamp the coordinates to [-1,1]
    m_minImSpaceX = (m_minImSpaceX < -1.0) ? -1.0 : m_minImSpaceX;
    m_minImSpaceY = (m_minImSpaceY < -1.0) ? -1.0 : m_minImSpaceY;
    m_maxImSpaceX = (m_maxImSpaceX > 1.0) ? 1.0 : m_maxImSpaceX;
    m_maxImSpaceY = (m_maxImSpaceY > 1.0) ? 1.0 : m_maxImSpaceY;
    
    //In case the coordinates are unlikely
    if(fabs(m_minImSpaceX - m_maxImSpaceX) < 0.000001)
    {
        m_minImSpaceX = -1.0;
        m_maxImSpaceX = 1.0;
    }
    
    if(fabs(m_minImSpaceY - m_maxImSpaceY) < 0.000001)
    {
        m_minImSpaceY = -1.0;
        m_maxImSpaceY = 1.0;
    }
}

/*! \brief Calculates scaling matrix.
 *
 *  Results are stored in m_mRange.
 */
void OVProjector::m_CalcMRange()
{
	float dx = m_maxImSpaceX-m_minImSpaceX;
	float dy = m_maxImSpaceY-m_minImSpaceY;
	m_mRange = OVm4(dx, 0, 0, m_minImSpaceX,
					 0, dy, 0, m_minImSpaceY,
					 0,  0, 1, 0,
					 0,  0, 0, 1);
    //Transpose since the shader uniform matrices are set to column major
    m_mRange = m_mRange.GetTranspose();
}

/*! \brief Initiates the projector
 *
 *  Stores pointers for camera primitives.
 *
 *  \param camPos Pointer to position of the camera
 *  \param camDir Pointer to direction of the camera
 *  \param camUp Pointer to up vector of the camera
 *  \param maxDisplacement Pointer to the maximum amplitude of waves, 
 *         used for volume calculations
 *  \param basePlaneY Enables vertical positioning of water surface 
 *  \param zoomFactor Pointer to the zoom factor of the camera 
 *  \param fovY Pointer to the field of view of the camera
 *  \param fovRatio Pointer to the ratio of height and width
 *  \param zNear Pointer to the distance from the camera to the far clip plane
 *  \param zFar Pointer to the distance from the camera to the near clip plane
 *  \param windowWidth Pointer to the window width 
 *  \param windowHeight Pointer to the window height
 */
void OVProjector::InitProjector(float *camPos,float *camDir,float *camUp, 
                                float* maxDisplacement, float basePlaneY, 
                                float* zoomFactor, float* fovY, float* fovRatio, 
                                float* zNear, float* zFar, int* windowWidth, 
                                int* windowHeight)
{
    m_camPos = camPos;
    m_camDir = camDir;
    m_camUp = camUp;
    m_maxDisplacement = maxDisplacement;
    m_baseplaneY = basePlaneY;

    m_zoomFactor = zoomFactor;
    m_fovY = fovY;
    m_fovRatio = fovRatio;
    m_zNear = zNear; 
    m_zFar = zFar;
    m_windowWidth = windowWidth;
    m_windowHeight = windowHeight;
}

/*! \brief Sets up the projector
 *
 *  Create the projector frustum using the camera's modelview matrix. 
 *  Save the current state, and then set the modelview matrix according to 
 *  the projector. Then construct the projector matrix, and finally set the 
 *  modelview matrix according to the camera again.
 */
void OVProjector::SetUpProjector()
{
    //Set up the projector frustum using the camera
    m_CalcProjectionFrustum();
    
    //Set the modelview matrix according to the projector
    Look();
    
    //Set the perspective (actually projection) matrix according 
    //to the projector
    Perspective();

    //Create the range matrix
    m_CalcFrustumWaterIntersections();
    m_CalcImSpaceMinMax();
    m_CalcMRange();
    
    //Get back to the camera's modelview matrix
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();    

    //Get back to the camera's projection matrix
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();    
}

/*! \brief Draws the projector frustum.
 *
 *  Draws the frustum using its cornerpoints.
 */
void OVProjector::DrawFrustum()
{
    //Set the modelview matrix according to the projector
    Look();

    //Get the current modelview matrix (the projector's)
    GLdouble modelView[16]; 
    glGetDoublev(GL_MODELVIEW_MATRIX, modelView);  
    
    //Set the perspective (actually projection) matrix according to 
    //the projector       
    Perspective();

    //Get the current projection matrix (the projector's)
    GLdouble projection[16]; 
    glGetDoublev(GL_PROJECTION_MATRIX, projection);    

    //Get the current viewport matrix
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    
    //Transform the frustum cornerpoints on the far clip plane 
    //to world coordinates
    OVv4 p1WorldSpace;
    GLdouble px=-100.0, py=10.0, pz=100.0;
    gluUnProject(0.0, 0.0, 1.0, modelView, projection, viewport,
                 &px, &py, &pz);
    p1WorldSpace[0] = px;
    p1WorldSpace[1] = py;
    p1WorldSpace[2] = pz;
    p1WorldSpace[3] = 1.0;     

    OVv4 p2WorldSpace;  
    gluUnProject(0.0, (GLdouble)(*m_windowHeight), 1.0, modelView, projection, 
                 viewport, &px, &py, &pz);
    p2WorldSpace[0] = px;
    p2WorldSpace[1] = py;
    p2WorldSpace[2] = pz;
    p2WorldSpace[3] = 1.0;

    OVv4 p3WorldSpace;  
    gluUnProject((GLdouble)(*m_windowWidth), (GLdouble)(*m_windowHeight), 1.0, 
                   modelView, projection, viewport, &px, &py, &pz);
    p3WorldSpace[0] = px;
    p3WorldSpace[1] = py;
    p3WorldSpace[2] = pz;
    p3WorldSpace[3] = 1.0;

    OVv4 p4WorldSpace;
    gluUnProject((GLdouble)(*m_windowWidth), 0.0, 1.0, modelView, projection, 
                    viewport, &px, &py, &pz);    
    p4WorldSpace[0] = px;
    p4WorldSpace[1] = py;
    p4WorldSpace[2] = pz;
    p4WorldSpace[3] = 1.0;

    //Transform the frustum cornerpoints on the near clip plane to world 
    //coordinates
    OVv4 p5WorldSpace;    
    gluUnProject(0.0, 0.0, 0.0, modelView, projection, viewport,
                 &px, &py, &pz);
    p5WorldSpace[0] = px;
    p5WorldSpace[1] = py;
    p5WorldSpace[2] = pz;
    p5WorldSpace[3] = 1.0;     

    OVv4 p6WorldSpace;  
    gluUnProject(0.0, (GLdouble)(*m_windowHeight), 0.0, modelView, projection, 
                 viewport, &px, &py, &pz);
    p6WorldSpace[0] = px;
    p6WorldSpace[1] = py;
    p6WorldSpace[2] = pz;
    p6WorldSpace[3] = 1.0;

    OVv4 p7WorldSpace;  
    gluUnProject((GLdouble)(*m_windowWidth), (GLdouble)(*m_windowHeight), 0.0, 
                 modelView, projection, viewport, &px, &py, &pz);
    p7WorldSpace[0] = px;
    p7WorldSpace[1] = py;
    p7WorldSpace[2] = pz;
    p7WorldSpace[3] = 1.0;

    OVv4 p8WorldSpace;
    gluUnProject((GLdouble)(*m_windowWidth), 0.0, 0.0, modelView, projection, 
                 viewport, &px, &py, &pz);
    p8WorldSpace[0] = px;
    p8WorldSpace[1] = py;
    p8WorldSpace[2] = pz;
    p8WorldSpace[3] = 1.0;

    //Restore the camera's modelview matrix
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    //Restore the camera's perspective (actually projection) matrix
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    //draw lines from the projector to the frustum corner points
    glColor3f(0.0, 1.0, 0.0);
    glBegin(GL_LINES);

    glVertex3f(m_projPos[0], m_projPos[1], m_projPos[2]);
    glVertex3f(p1WorldSpace[0], p1WorldSpace[1], p1WorldSpace[2]);

    glVertex3f(m_projPos[0], m_projPos[1], m_projPos[2]);
    glVertex3f(p2WorldSpace[0], p2WorldSpace[1], p2WorldSpace[2]);

    glVertex3f(m_projPos[0], m_projPos[1], m_projPos[2]);
    glVertex3f(p3WorldSpace[0], p3WorldSpace[1], p3WorldSpace[2]);
    
    glVertex3f(m_projPos[0], m_projPos[1], m_projPos[2]);
    glVertex3f(p4WorldSpace[0], p4WorldSpace[1], p4WorldSpace[2]);

    glEnd();

    //draw the far plane
    glBegin(GL_LINE_LOOP);
    
    glVertex3f(p1WorldSpace[0], p1WorldSpace[1], p1WorldSpace[2]);
    glVertex3f(p2WorldSpace[0], p2WorldSpace[1], p2WorldSpace[2]);
    glVertex3f(p3WorldSpace[0], p3WorldSpace[1], p3WorldSpace[2]);
    glVertex3f(p4WorldSpace[0], p4WorldSpace[1], p4WorldSpace[2]);

    glEnd();

    //draw the near plane
    glBegin(GL_LINE_LOOP);
    
    glVertex3f(p5WorldSpace[0], p5WorldSpace[1], p5WorldSpace[2]);
    glVertex3f(p6WorldSpace[0], p6WorldSpace[1], p6WorldSpace[2]);
    glVertex3f(p7WorldSpace[0], p7WorldSpace[1], p7WorldSpace[2]);
    glVertex3f(p8WorldSpace[0], p8WorldSpace[1], p8WorldSpace[2]);

    glEnd();   	
}

/*! \brief Sets the modelview matrix according to the projector.
 *
 *  Sets the modelview matrix according to the projector position, 
 *  direction and up vectors by calling gluLookAt. Allows the previous
 *  modelview matrix to be restored by pushing it on the stack.
 */
void OVProjector::Look()
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();    
    
    glLoadIdentity();
    
    //Use the projector's modelview matrix
    gluLookAt(m_projPos[0], m_projPos[1], m_projPos[2], 
        m_projPos[0] + m_projDir[0], m_projPos[1] + m_projDir[1], 
        m_projPos[2] + m_projDir[2], m_projUp[0], m_projUp[1], m_projUp[2]);
}

/*! \brief Sets the projection matrix according to the projector.
 *
 *  Sets the projection matrix according to the projector zoomfactor, 
 *  field of view, field of view ratio, far plane, near plane, and the 
 *  window width and window height by calling gluPerspective. Allows the
 *  previous projection matrix to be restored by pushing it on the stack.
 *
 */
void OVProjector::Perspective()
{ 
    int height = (*m_windowHeight <= 0) ? 1 : *m_windowHeight;
    int width = *m_windowWidth;

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();  
    
    glLoadIdentity();
    
    //Increase the field of view with 35 degrees
    //compared to the camera's, to make sure  
    //that the projected mesh covers the whole
    //water surface
    //Use the projector's perspective (actually projection) matrix
    gluPerspective((*m_zoomFactor) * ((*m_fovY)+35.0), 
                    (*m_fovRatio)*((float)width)/((float)height), 
                    (*m_zNear), 
                    (*m_zFar) ); 

}

/*! \brief Indicates whether water should be drawn or not.
 *
 *  If the projector frustum intersects the lower or upper displacement plane
 *  the water is visible to the viewer, otherwise it isn't and then it's 
 *  unnecessairy to draw it.
 *  
 *  \return True if water is visible for the viewer, false otherwise.
 */
bool OVProjector::Visible()
{ 
    return !m_intersectionsWorldSpace.empty();
}

}  // End of OV
