/*! \file ovprojector.h
 *  \brief OVProjector header file.
 *
 *  Contains the declarations of the OVProjector class. 
 */

#ifndef _OVPROJECTOR_H_
#define _OVPROJECTOR_H_

#include <vector>
#include <GL/glew.h>
#include <GL/glut.h>
#include "ovalgebra.h"

namespace OV
{

/*! \brief The OVProjector class.
 *  
 *  Calculates MRange and MProjection matrices by calculating projection
 *  frustom from camera primitives. Its main objective is to create a projection
 *  matrix from image space to water plane in world space.
 */
class OVProjector
{
public:
    OVProjector();
    ~OVProjector();

    void InitProjector(float *camPos,float *camDir,float *camUp, 
                       float* maxDisplacement, float basePlaneY, 
                       float* zoomFactor, float* fovY, float* fovRatio, 
                       float* zNear, float* zFar, int* windowWidth, 
                       int* windowHeight);
    void SetUpProjector();
    float* GetMRange();
    void DrawFrustum();
    void Look();
    void Perspective();
    bool Visible();

protected:
    void m_CalcProjectionFrustum();
    void m_CalcFrustumWaterIntersections();
    void m_CalcImSpaceMinMax();
    void m_CalcMRange();
    bool m_CalcWaterplaneYLineIntersection(float planeY, OVv4 point1,
                                           OVv4 point2, OVv4* intersection);

private:
    float *m_zoomFactor; /*!< Camera zoom factor (pointer to shell)*/
    float *m_fovY;       /*!< Camera FOV for y-axis (pointer to shell)*/    
    float *m_fovRatio;   /*!< Camera FOV ratio (pointer to shell)*/
    float *m_zNear;      /*!< Camera near plane (pointer to shell)*/
    float *m_zFar;       /*!< Camera far position (pointer to shell)*/
    int* m_windowWidth;  /*!< The window width (pointer to shell)*/
    int* m_windowHeight; /*!< The window height (pointer to shell)*/
    
    float *m_camPos;  /*!< Camera position (pointer to shell) */
    float *m_camDir;  /*!< Camera direction (pointer to shell) */    
    float *m_camUp;   /*!< Camera up vector (pointer to shell) */
    float *m_projPos; /*!< Projector position */
    float *m_projDir; /*!< Projector direction */
    float *m_projUp;  /*!< Projector up vector */

    std::vector<OVv4> m_intersectionsWorldSpace; /*!< Intersections between
                                                  * projector frustum and water
                                                  * volume in world space
                                                  * coordinates */
    std::vector<OVv4> m_intersectionsImageSpace; /*!< Intersections between
                                                  * projector frustum and water
                                                  * volume when projected to 
                                                  * image space coordinates */

    float* m_maxDisplacement; /*!< Maximum displacement of the water surface
                               * dependent on windspeed etc.
                               * (pointer to ovwater)*/
    float m_baseplaneY;  /*!< Offset from Y=0 in world space coordinates */
    OVm4 m_mRange;       /*!< Matrix for scaling grid in image space, caculated
                          * from min max coverage values */
    float m_minImSpaceX; /*!< Water min x coverage in image space */
    float m_minImSpaceY; /*!< Water min y coverage in image space */
    float m_maxImSpaceX; /*!< Water max x coverage in image space */
    float m_maxImSpaceY; /*!< Water max y coverage in image space */ 
};

}  // End of OV

#endif // End of _OVPROJECTOR_H_
