/*! \file shader.cpp
 *  \brief Shader handling source file.
 *
 *  This file contains the methods for the shader object which is used to
 *  compile and run shaders.
 */

#include "ovshader.h"

namespace OV
{

/*! \brief Constructor for the shader class.
 *
 * Constructor for the shader class, resets all variables.
 */
OVShader::OVShader()
{
    m_vertex = 0;
	m_fragment = 0;
	m_program = 0;
	m_name = 0;
}

/*! \brief Constructor for the shader class.
 *
 *  Constructor for the shader class, loads and compiles a specified shader.
 *
 *  \param name Name of the shader program.
 */
OVShader::OVShader(const char* name)
{
    m_vertex = 0;
	m_fragment = 0;
	m_program = 0;

    Compile(name);
}

/*! \brief Destructor for the shader class.
 *
 *  Releases any loaded shaderprogram.
 */
OVShader::~OVShader()
{
    glDeleteShader(m_vertex);
    glDeleteShader(m_fragment);
    glDeleteProgram(m_program);
    delete[] m_name;
	for(int i = 0; i < (int)m_uniforms.size(); i++)
        delete m_uniforms[i];
}

/*! \brief Compile a shader program.
 *
 *  Read a vertex and a fragment program from source and compile it.
 *
 *  \param shaderName Name of the shader program.
 */
void OVShader::Compile(const char* shaderName)
{
    if(m_name != 0)
        delete[] m_name;
    m_name = new char[std::strlen(shaderName) + 1];
	std::strcpy(m_name, shaderName);

    char vSource[40];
    sprintf(vSource, "%s.vert", m_name);
    char* vCode = Txt2char(vSource);
     
    m_vertex = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(m_vertex, 1, (const GLchar**)&vCode, NULL);
    glCompileShader(m_vertex);
    free(vCode);
     
    GLint vertexCompiled;
    glGetShaderiv(m_vertex, GL_COMPILE_STATUS, &vertexCompiled);
    if(!vertexCompiled)
    {
        GLint vLength;
        GLchar* vLog;
        glGetShaderiv(m_vertex, GL_INFO_LOG_LENGTH, &vLength);
        vLog = (GLchar*) malloc(vLength);
        glGetShaderInfoLog(m_vertex, vLength, &vLength, vLog);
        std::cout<<"Vertex shader compile log= "<<vLog<<std::endl;              
    }
     
    char fSource[40];
    sprintf(fSource, "%s.frag", m_name);
    char* fCode = Txt2char(fSource);
     
    m_fragment = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(m_fragment, 1, (const GLchar**)&fCode, NULL);
    glCompileShader(m_fragment);
    free(fCode);

    GLint fragmentCompiled;
    glGetShaderiv(m_fragment, GL_COMPILE_STATUS, &fragmentCompiled);
    if(!fragmentCompiled)
    {
        GLint fLength;
        GLchar* fLog;
        glGetShaderiv(m_fragment, GL_INFO_LOG_LENGTH, &fLength);
        fLog = (GLchar*) malloc(fLength);
        glGetShaderInfoLog(m_fragment, fLength, &fLength, fLog);
        std::cout<<"Fragment shader compile log= "<<fLog<<std::endl;              
    }
          
    m_program = glCreateProgram();
     
    glAttachShader(m_program, m_vertex);
    glAttachShader(m_program, m_fragment);
     
    glLinkProgram(m_program);
     
    GLint programLinked;
    glGetProgramiv(m_program, GL_LINK_STATUS, &programLinked);
    if(!programLinked)
    {
        GLint pLength;
        GLchar* pLog;
        glGetProgramiv(m_program, GL_INFO_LOG_LENGTH, &pLength);
        pLog = (GLchar*) malloc(pLength);
        glGetShaderInfoLog(m_program, pLength, &pLength, pLog);
        std::cout<<"Program linking log= "<<pLog<<std::endl;              
    } 
}

/*! \brief Update all uniform variables.
 *
 *  Update all uniform variables that has been connected to this
 *  shader program.
 */
void OVShader::UpdateUniforms()
{
    std::vector<OVUniVar*>::iterator uniVarIterator;
    for(uniVarIterator = m_uniforms.begin(); 
        uniVarIterator != m_uniforms.end(); uniVarIterator++)
    {
        (*uniVarIterator)->Update();
    }
}

/*! \brief Update locations for all uniform variables.
 *
 *  Update locations for all uniform variables that has been connected to this
 *  shader program.
 */
void OVShader::UpdateUniformsLocation()
{
    std::vector<OVUniVar*>::iterator uniVarIterator;
    for(uniVarIterator = m_uniforms.begin(); 
        uniVarIterator != m_uniforms.end(); uniVarIterator++)
    {
        (*uniVarIterator)->UpdateLocation(m_program);
    }  
}

/*! \brief Activate shader program.
 *
 *  Set this proram as the active shader program.
 */
void OVShader::Activate()
{
    glUseProgram(m_program);
}

/*! \brief Deactivate shader program.
 *
 *  Set the active shader program to 0.
 */
void OVShader::Deactivate()
{
    glUseProgram(0);
}

/*! \brief Add uniform variable.
 *
 *  Add a uniform variable to this shader. The added variables
 *  will then be updated automaticly by the shader.
 */
void OVShader::AddUniform(OVUniVar* uniVar)
{
    m_uniforms.push_back(uniVar);
}

/*! \brief Read shader program from file.
 *
 *  This method opens a file and reads it into a char-buffer.
 *
 *  \param fileName Filename to read.
 *  \return char-buffer with the file content.
 */
char* OVShader::Txt2char(const char* fileName)
{
	FILE* textFile;
	char* fileContent = NULL;
	int fileLength = 0;
	
    if (fileName != NULL) 
    {
		textFile = fopen(fileName, "rt");
		if (textFile != NULL) 
        {
			fseek(textFile, 0, SEEK_END);
			fileLength = ftell(textFile);
			rewind(textFile);
			if (fileLength > 0) 
            {
				fileContent = (char *)malloc(sizeof(char) * (fileLength+4));
				fileLength = (int)fread(fileContent, sizeof(char), fileLength,
                                        textFile);
				fileContent[fileLength] = '\0';
			}
			fclose(textFile);
		}
		else 
        {
            std::cout<<"The file could not be opened (fopen = NULL!)"
                     <<std::endl;
	    }
    }
	else
    {
        std::cout<<"The file could not be read from (fname = NULL!)"
                 <<std::endl;
    }
    return fileContent;
}

}  // End of OV
