/*! \file univar.h
 *  \brief Uniform variable header file.
 *
 *  This file contains an object for handling uniform variables.
 */

#ifndef _OVUNIVAR_H_
#define _OVUNIVAR_H_

#include <cstring>
#include <GL/glew.h>

namespace OV
{

/*! \brief Class that handles uniform variables.
 *
 *  This class is only used as base class for the real UniVar classes.
 *  Templates can not be used because different GL-functions are used
 *  to pass parameters to GLSL.
 */
class OVUniVar
{
public:
    OVUniVar();
    virtual ~OVUniVar();
    virtual void UpdateLocation(GLuint program);
    virtual void Update();

protected:
    char* m_name; /*!< Name of the shader program */
    GLint m_location; /*!< Location of the variable whithin the program */
};

/*! \brief Class that handles uniform integer variables.
 *
 *  This class encapsulates uniform integer variables used in shaders. 
 *  It inherits from OVUniVar. 
 */
class OVUniVari : public OVUniVar
{
public:
    OVUniVari(const char* name, int* source);
    void Update();
private:
    int* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float variables.
 *
 *  This class encapsulates uniform float variables used in shaders. 
 *  It inherits from OVUniVar. 
 */
class OVUniVarf : public OVUniVar
{
public:
    OVUniVarf(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float vector variables of length 2.
 *
 *  This class encapsulates uniform float vector variables 
 *  of length 2 used in shaders. It inherits from OVUniVar. 
 */
class OVUniVar2fv : public OVUniVar
{
public:
    OVUniVar2fv(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float vector variables of length 3.
 *
 *  This class encapsulates uniform float vector variables 
 *  of length 3 used in shaders. It inherits from OVUniVar. 
 */
class OVUniVar3fv : public OVUniVar
{
public:
    OVUniVar3fv(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float matrix variables of dimension 3x3.
 *
 *  This class encapsulates uniform float matrix variables 
 *  of dimension 3x3 used in shaders. It inherits from OVUniVar. 
 */
class OVUniVar9fm : public OVUniVar
{
public:
    OVUniVar9fm(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float vector variables of length 4.
 *
 *  This class encapsulates uniform float vector variables 
 *  of length 4 used in shaders. It inherits from OVUniVar. 
 */
class OVUniVar4fv : public OVUniVar
{
public:
    OVUniVar4fv(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

/*! \brief Class that handles uniform float matrix variables of dimension 4x4.
 *
 *  This class encapsulates uniform float matrix variables 
 *  of dimension 4x4 used in shaders. It inherits from OVUniVar. 
 */
class OVUniVar16fm : public OVUniVar
{
public:
    OVUniVar16fm(const char* name, float* source);
    void Update();
private:
    float* m_source; /*!< Pointer to the data source */
};

}  // End of OV

#endif // End of _OVUNIVAR_H_
