/*! \file globals.h
 *  \brief Declarations of global variables
 */

#ifndef _GLOBALS_H_
#define _GLOBALS_H_

#include <GL/glew.h>
#include "vec3.h"
#include "camera.h"
#include "timer.h"
#include "mesh.h"
#include "shader.h"

//
// Misc
//

// Environment settings
extern GLfloat sunPosition[4];  /*!< Position of the sun in world 
                                     coordinates */
extern GLfloat sunDirection[3]; /*!< Direction of the sunlight in world 
                                     coordinates */
extern GLfloat sunColor[4];     /*!< Diffuse and specular color of the sun */
extern GLfloat sunAmbColor[4];  /*!< Ambient color of the sun */
extern GLuint sunLightSource;   /*!< Light source for the sun */
extern float sunAngle;          /*!< Sun angle [0 pi] */
extern float windVelocity[2];   /*!< Wind velocity */
extern float windDir;           /*!< Wind direction */
extern float windSpeed;         /*!< Wind speed */
extern float waterColor[3];     /*!< Color of the water */

// Camera
extern GLfloat cameraPosition[4];  /*!< Camera position in world coordinates */
extern GLfloat cameraDirection[4]; /*!< Camera direction in world coordinates */
extern GLfloat cameraUpVector[4];  /*!< Camera up-vector in world coordinates */

extern GLfloat zoomFactor; /*!< The amount of zooming */  
extern GLfloat fovY; /*!< The field-of-view angle in the y direction */
extern GLfloat fovRatio; /*!< The aspect ratio of the field of view (x/y) */
extern GLfloat zNear; /*!< The distance from the camera to the near clipping plane */
extern GLfloat zFar; /*!< The distance from the camera to the far clipping plane */

// Water settings
extern float waveAmpFactor; /*!< Controls the amplitude of the waves */

// Time
extern Timer timer;   /*!< Timer for the system */
extern float t;       /*!< Time variable to use in animations */
extern float dt;      /*!< Time elapsed since last frame to use in animations */
extern float daytime; /*!< Time of day [0 1] -> 00:00-24:00. */

// Quality
extern int lodRes; /*!< Level of detail for grid */
extern int lodVis; /*!< Level of detail for visualisation shader */

// Ground
extern Mesh groundMesh;               /*!< Mesh for the ground */
extern Shader groundShader;           /*!< Shader for the ground */
extern Texture2d groundNoiseTexture;  /*!< Texture for the ground */
extern Texture2d groundHeightTexture; /*!< Height texture for the ground */
extern Texture2d groundTexture;       /*!< Texture for the ground */

// Sky
extern Texture2d skyTexture;      /*!< Texture for the sky */
extern Texture2d skyCloudTexture; /*!< Texture for the clouds */
extern Shader skyShader;          /*!< Shader for colorizing the sky */
extern float skyHeight;           /*!< Help variable for skyShader */
extern float skyOffset;           /*!< Help variable for skyShader */

// Boat
extern Texture2d boatSideTexture; /*!< Texture for the side of the boat */
extern Texture2d boatDeckTexture; /*!< Texture for the deck of the boat */
extern Texture2d boatTopTexture;  /*!< Texture for the top of the boat */

// Render settings
extern GLuint polygonMode; /*!< Variable for drawing wirefram or solid */

// 
// Window management
//

extern int windowID;          /*!< ID to the Glut window */
extern int windowWidth;       /*!< Width of the Glut window */
extern int windowHeight;      /*!< Height of the Glut window */
extern Texture2d helpTexture; /*!< Texture for the help menu */

//
// Keyboard input
//

// Camera movement
extern Camera camera;
extern float moveSpeed;     /*!< Speed for movement of the camera */
extern float turnSpeed;     /*!< Angular velocity for camera turning in deg/s */
extern float zoomSpeed;     /*!< Zoom speed for camera 1.0 means that full zoom
                                 in or out from initial zoom takes 1 second */
extern int keyboardStrafe;  /*!< Camera strafe left: -1, stop: 0, right: 1 */
extern int keyboardForward; /*!< Camera forward move: 1, stop: 0, 
                                 backwards: -1 */
extern int keyboardYaw;     /*!< Camera yaw left: -1, stop: 0, right: 1 */
extern int keyboardRoll;    /*!< Camera roll left: -1, stop: 0, right: 1 */
extern int keyboardPitch;   /*!< Camera pitch down: -1, stop: 0, up: 1 */
extern int keyboardZoom;    /*!< zoom in: 1, zoom out: -1, don't zoom: 0 */
extern int keyboardElevate; /*!< Up: 1, down: -1, stop: 0 */


// Environment
extern float sunSpeed;      /*!< Turnspeed for the sun */
extern float waterHorizon;  /*!< Distance to the water horizon */
extern int keyboardMoveSun; /*!< Earlier: 1, Later: -1, stop: 0 */
extern float windMoveSpeed;  /*!< Speed for changing wind velocity */
extern int keyboardMoveWind; /*!< Less wind: -1, More wind: 1, No change: 0 */
extern int keyboardTurnWind; /*!< CW: -1, CCW: 1, No change: 0 */

// Flags
extern bool keyboardDrawWater;   /*< Draw water */
extern bool keyboardDrawSky;     /*< Draw sky */
extern bool keyboardDrawTerrain; /*< Draw terrain */
extern bool keyboardDrawBoat;    /*< Draw floating boat */
extern bool keyboardLook;        /*!< Camera look at frustum */
extern bool keyboardHelp;        /*!< Show help menu */
extern bool keyboardParameters;  /*!< Show parameters */
extern bool keyboardFPS;         /*!< Show time and FPS */
extern bool keyboardAxis;        /*!< Show axis */
extern bool keyboardWind;        /*!< Show wind direction */
extern bool keyboardLight;       /*!< Show light direction */

#endif // End of _GLOBALS_H_
