/*! \file main.cpp
 *  \brief Main source file.
 *
 *  Startup source file for the OceanView project.
 */

#include <iostream>
#include <GL/glew.h>
#include <GL/glut.h>
#include "globals.h"
#include "display.h"
#include "input.h"
#include "ovwater.h"
#include "mesh.h"
#include "shader.h"
#include "univar.h"
#include "texture2d.h"

/*! \brief Reshapes the Glut window
 *
 *  This function is called when the Glut window is resized. It sets the new
 *  viewport size and tells the camera to change the projection matrix.
 *
 *  \param w new width.
 *  \param h new height.
 */
void Reshape(int w, int h)
{
    windowWidth = w;
    windowHeight = h;
    camera.Perspective(w, h);
    camera.GetCameraVariables(cameraPosition, cameraDirection, cameraUpVector,
                              zoomFactor, fovY, fovRatio, zNear, zFar);
}

/*! \brief Idle function in the Glut loop
 *
 *  This function is called when a frame is rendered. It handles input
 *  and calculates the new time and delta time.
 */ 
void Idle()
{
    UpdateInput();
    dt = (float)timer.GetDeltaTime();
    t += dt;
    glutPostRedisplay();
}

/*! \brief Display function for Glut
 *
 *  This function draws a frame and is called back again using 
 *  glutPostRedisplay. It clears the framebuffer, calls the
 *  DrawScene function and swaps buffers.
 */
void Display()
{
   	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    camera.Look();

    // Update light source
    glLightfv(sunLightSource, GL_POSITION, sunPosition);
    glLightfv(sunLightSource, GL_SPOT_DIRECTION, sunDirection);
    glLightfv(sunLightSource, GL_AMBIENT, sunAmbColor);
    glLightfv(sunLightSource, GL_SPECULAR, sunColor);
    glLightfv(sunLightSource, GL_DIFFUSE, sunColor);

    // Set polygon mode
    glPolygonMode(GL_FRONT_AND_BACK, polygonMode);

    // Draw the scene (io/display.cpp)
    DrawScene();

    glutSwapBuffers();
}

/*! \brief Initialization function for OpenGL
 *
 *  This function sets up everything in Glut and OpenGL that is
 *  needed. Glew is also initialized from here. This must be run
 *  before textures or shaders are loded.
 */
void InitGL()
{
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutInitWindowSize(windowWidth, windowHeight);
    windowID = glutCreateWindow("OceanView");

    GLenum err = glewInit();

    if (GLEW_OK != err) 
    {
        std::cerr << "GLEW error: " << glewGetErrorString(err) << std::endl;
        exit(0);
    }

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
	glEnable(GL_POINT_SMOOTH);
	glShadeModel(GL_SMOOTH);

    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glClearColor(0.0,0.0,0.0,1.0);

    glutDisplayFunc(Display);
    glutReshapeFunc(Reshape);
    glutIdleFunc(Idle);

    glutIgnoreKeyRepeat(GL_TRUE);
    glutKeyboardFunc(KeyDownFunc);
    glutKeyboardUpFunc(KeyUpFunc);
    glutSpecialFunc(SpecDownFunc);
    glutSpecialUpFunc(SpecUpFunc);
}

/*! \brief Load textures
 * 
 *  Load textures for the testshell.
 */
void LoadTextures()
{
    groundTexture.Load("testshell/textures/island.tga");
    groundHeightTexture.Load("testshell/textures/island_height.tga", true);
    groundNoiseTexture.Load("testshell/textures/noise.tga");

    helpTexture.Load("testshell/textures/help.tga");

    skyTexture.Load("testshell/textures/clearSky.tga");
    skyCloudTexture.Load("testshell/textures/clouds.tga");

    boatSideTexture.Load("testshell/textures/boat_side.tga");
    boatDeckTexture.Load("testshell/textures/boat_deck.tga");
    boatTopTexture.Load("testshell/textures/boat_top.tga");
}

/*! \brief Load shaders
 *
 *  Shaders are loaded in this function, also uniform variables
 *  can be created here.
 */
void LoadShaders()
{
    // Constants for texture units
    static int texUnitIsland = 0;
    static int texUnitNoise = 1;
    static int texUnitSky = 0;
    static int texUnitClouds = 1;

    groundShader.Compile("testshell/shaders/ground");
    groundShader.AddUniform(new UniVari("tex_island", &texUnitIsland));
    groundShader.AddUniform(new UniVari("tex_noise", &texUnitNoise));
  
    groundShader.Activate();
    groundShader.UpdateUniformsLocation();
    groundShader.Deactivate();

    skyShader.Compile("testshell/shaders/sky");
    skyShader.AddUniform(new UniVari("texSky", &texUnitSky));
    skyShader.AddUniform(new UniVari("texClouds", &texUnitClouds));
    skyShader.AddUniform(new UniVarf("daytime", &daytime));
    skyShader.AddUniform(new UniVarf("height", &skyHeight));
    skyShader.AddUniform(new UniVarf("offset", &skyOffset));

    skyShader.Activate();
    skyShader.UpdateUniformsLocation();
    skyShader.Deactivate();
}

/*! \brief Initialize scene
 *
 */
void InitScene()
{
    groundMesh.LoadHeightMap("testshell/textures/island_height.tga",
                             -100, 100, -100, 100, -15, 15);
    groundMesh.CreateVBO();
}

/*! \brief The main function for the testshell 
 *
 *  This is the main function in the testshell. It starts Glut and initializes
 *  OceanView.
 *
 *  \param argc argument count.
 *  \param argv arguments.
 *  \return result.
 */
int main(int argc, char** argv)
{
    std::cout<<"Starting testshell"<<std::endl;
	glutInit(&argc, argv);

    std::cout<<"Initializing OpenGL"<<std::endl;
    InitGL();

    std::cout<<"Initializing scene"<<std::endl;
    InitScene();
  
    std::cout<<"Loading textures"<<std::endl;
    LoadTextures();
    
    std::cout<<"Loading shaders"<<std::endl;
    LoadShaders();

    std::cout<<"Loading OVWater"<<std::endl;
    OV::OVInitWater(cameraPosition, cameraDirection, cameraUpVector, &t, 
                    &daytime, &waterHorizon, windVelocity, waterColor, sunColor, 
                    lodRes, lodVis, &skyCloudTexture.texID,
                    &skyTexture.texID, &groundHeightTexture.texID,
                    &zoomFactor, &fovY, &fovRatio, &zNear, &zFar, 
                    &windowWidth, &windowHeight);

    std::cout<<"Up and running"<<std::endl;

    glutMainLoop();

    return 0;
}
