/*! \file camera.cpp
 *  \brief Camera file.
 *
 *  Contains the definitions of the Camera class. The camera's orientation and
 *  position can be altered. The camera can be used to specify a modelview 
 *  matrix and a projection matrix. The camera frustum can be drawn on the 
 *  screen.
 */

#include "camera.h"

/*! \brief The constructor of the camera class
 *
 *  Initializes all variables.
 */
Camera::Camera()
{
    m_pos = Vec3f(0.0f, 20.0f, -50.0f);
    m_up = Vec3f(0.0f, 1.0f, 0.0f);
    m_viewDir = Vec3f(0.0f, 0.0f, 1.0f);
    m_right = Vec3f(1.0f, 0.0f, 0.0f);   
    
    m_yaw = 0.0f;
    m_pitch = 0.0f;
    m_roll = 0.0f;

    m_fovY = 45.0f;
    m_fovRatio = 1.0f;
    m_fovX = m_fovY * m_fovRatio;
    m_zNear = 1.0f;
    m_zFar = 10000.0f;
    m_zoomFactor = 1.0f;

    m_frustumView = false;

    SetState(0, 300, -700, 0, -18, 0, 45, 1, m_zNear, m_zFar, 1);
}

/*! \brief The destructor of the camera class
 *
 */
Camera::~Camera() { }

/*! \brief Sets the state of the camera
 *  
 *  Sets the position, yaw, pitch, roll, FOV in the Y-direction, FOV-ratio, 
 *  as well as the distance to the near clip plane and the far clip plane.
 *  
 *  \param x the x position of the camera.
 *  \param y the y position of the camera.
 *  \param z the z position of the camera.
 *  \param yaw the amount of yaw in degrees.
 *  \param pitch the amount of pitch in degrees.
 *  \param roll the amount of roll in degrees.
 *  \param fovY the angle of the FOV in degrees.
 *  \param fovRatio the ratio of the FOV (fovX/fovY).
 *  \param zNear the distance from the camera to the near clip plane.
 *  \param zFar the distance from the camera to the far clip plane.
 *  \param zoomFactor Zoomfactor, 1.0 means no zoom.
 */
void Camera::SetState(GLfloat x, GLfloat y, GLfloat z, GLfloat yaw, 
                      GLfloat pitch, GLfloat roll, GLfloat fovY, 
                      GLfloat fovRatio, GLfloat zNear, GLfloat zFar, 
                      GLfloat zoomFactor)
{
    m_pos = Vec3f(x, y, z);

    m_up = Vec3f(0.0f, 1.0f, 0.0f);
    m_viewDir = Vec3f(0.0f, 0.0f, 1.0f);
    m_right = Vec3f(1.0f, 0.0f, 0.0f);

    m_yaw = 0.0f;
    m_pitch = 0.0f;
    m_roll = 0.0f;

    m_Yaw(yaw);
    m_Pitch(pitch);
    m_Roll(roll);
    
    m_fovY = fovY;
    m_fovRatio = fovRatio;
    m_fovX = fovY * fovRatio;
    m_zNear = zNear;
    m_zFar = zFar;

    m_zoomFactor = zoomFactor;

}

/*! \brief Sets the viewing range of the camera
 *  
 *  Sets the near and far clip plane.
 *  
 *  \param zNear the distance from the camera to the near clip plane.
 *  \param zFar the distance from the camera to the far clip plane.
 */
void Camera::SetRange(GLfloat zNear, GLfloat zFar)
{
    m_zNear = zNear;
    m_zFar = zFar;
    Perspective(m_windowWidth, m_windowHeight);
}

/*! \brief Moves the camera up/down.
 *  
 *  Moves the camera up or down in the "up" direction.
 *  
 *  \param step the elevation step.
 */
void Camera::m_Elevate(GLfloat step)
{
    m_pos = m_pos + step*m_up;
}

/*! \brief Moves the camera forward/backwards.
 *  
*  Moves the camera forwards or backwards in the viewing direction.
 *  
 *  \param step the forward/backwards step.
 */
void Camera::m_Forward(GLfloat step)
{
    m_pos = m_pos + step*m_viewDir;
}

/*! \brief Moves the camera right/left.
 *  
 *  Moves the camera in the right/left direction.
 *  
 *  \param step the strafe step.
 */
void Camera::m_Strafe(GLfloat step)
{
    m_pos = m_pos + step*m_right;
}

/*! \brief Zooms the camera.
 *  
 *  Adjusts the zoom factor which is used in the cameras perspective function.
 *  
 *  \param zoomFactor the change of zoom factor.
 */
void Camera::m_Zoom(GLfloat zoomFactor)
{   
    m_zoomFactor = clamp(m_zoomFactor + zoomFactor, 0.35f, 1.5f);
}
	
/*! \brief Rotates the camera around the up vector.
 *  
 *  Rotates the camera around the up vector (yaw).
 *  
 *  \param angle the change of yaw angle (degrees).
 */
void Camera::m_Yaw(GLfloat angle)
{
    m_yaw = m_yaw + angle;
    angle = m_DegreesToRadians(angle);
    m_viewDir = m_viewDir*cos(angle) - m_right*sin(angle);
    m_viewDir.Normalize();
    m_right = m_viewDir.Cross(m_up); 
}

/*! \brief Rotates the camera around the right vector.
 *  
 *  Rotates the camera around the right vector (pitch).
 *  
 *  \param angle the change of pitch angle (degrees).
 */
void Camera::m_Pitch(GLfloat angle)
{
    m_pitch = m_pitch + angle;
    angle = m_DegreesToRadians(angle);    
  	m_viewDir = m_viewDir*cos(angle) + m_up*sin(angle);
    m_viewDir.Normalize();
    m_up = m_viewDir.Cross(m_right)*(-1.0f);
}

/*! \brief Rotates the camera around the view vector.
 *  
 *  Rotates the camera around the view vector (roll).
 *  
 *  \param angle the change of roll angle (degrees).
 */
void Camera::m_Roll(GLfloat angle)
{
    m_roll = m_roll + angle;		
    angle = m_DegreesToRadians(angle);
  	m_right = m_right*cos(angle) + m_up*sin(angle);
    m_right.Normalize();
  	m_up = -m_viewDir.Cross(m_right);
}

/*! \brief Draws the camera frustum.
 *  
 *  Draws a wireframe model of the cameras frustum. 
 */
void Camera::Draw()
{   
    glPushAttrib(GL_LIGHTING_BIT);
    glPushAttrib(GL_CURRENT_BIT);
    glColor3f(1.0, 0.0, 0.0);

    //normalize vectors
    m_viewDir = m_viewDir.Normalize();
    m_up = m_up.Normalize();
    m_right = m_right.Normalize();
    
    m_viewPt = m_pos + m_viewDir;
      
    //Calculate some distances
    GLfloat fovYRad = m_DegreesToRadians(m_zoomFactor * m_fovY);
    GLfloat fovXRad = m_DegreesToRadians(m_zoomFactor * m_fovX);
    GLfloat deltaYNear = tanf(fovYRad/2.0f) * m_zNear;
    GLfloat deltaXNear = tanf(fovXRad/2.0f) * m_zNear;
    GLfloat deltaYFar = tanf(fovYRad/2.0f) * m_zFar;
    GLfloat deltaXFar = tanf(fovXRad/2.0f) * m_zFar;

    //The vertices on the near clip plane
    Vec3f nearTopLeft = m_pos + m_zNear * m_viewDir + deltaYNear * m_up - 
                        deltaXNear * m_right;
    Vec3f nearTopRight = m_pos + m_zNear * m_viewDir + deltaYNear * m_up +
                         deltaXNear * m_right;
    Vec3f nearBottomLeft = m_pos + m_zNear * m_viewDir - deltaYNear * m_up -
                           deltaXNear * m_right;
    Vec3f nearBottomRight = m_pos + m_zNear * m_viewDir - deltaYNear * m_up +
                            deltaXNear * m_right;
    
    //The vertices on the far clip plane
    Vec3f farTopLeft = m_pos + m_zFar * m_viewDir + deltaYFar * m_up -
                       deltaXFar * m_right;
    Vec3f farTopRight = m_pos + m_zFar * m_viewDir + deltaYFar * m_up +
                        deltaXFar * m_right;
    Vec3f farBottomLeft = m_pos + m_zFar * m_viewDir - deltaYFar * m_up -
                          deltaXFar * m_right;
    Vec3f farBottomRight = m_pos + m_zFar * m_viewDir - deltaYFar * m_up +
                           deltaXFar * m_right;    
    
    //draw the near clip plane
    glBegin(GL_LINE_LOOP);    
    glVertex3f(nearTopLeft.x, nearTopLeft.y, nearTopLeft.z);
    glVertex3f(nearTopRight.x, nearTopRight.y, nearTopRight.z);
    glVertex3f(nearBottomRight.x, nearBottomRight.y, nearBottomRight.z);
    glVertex3f(nearBottomLeft.x, nearBottomLeft.y, nearBottomLeft.z);
    glEnd();
    
    //draw the far clip plane
    glBegin(GL_LINE_LOOP);
    glVertex3f(farTopLeft.x, farTopLeft.y, farTopLeft.z);
    glVertex3f(farTopRight.x, farTopRight.y, farTopRight.z);
    glVertex3f(farBottomRight.x, farBottomRight.y, farBottomRight.z);
    glVertex3f(farBottomLeft.x, farBottomLeft.y, farBottomLeft.z);
    glEnd();
    
    glBegin(GL_LINE_STRIP);
    glVertex3f(m_pos.x, m_pos.y, m_pos.z);
    glVertex3f(farTopLeft.x, farTopLeft.y, farTopLeft.z);

    glVertex3f(m_pos.x, m_pos.y, m_pos.z);
    glVertex3f(farTopRight.x, farTopRight.y, farTopRight.z);
    
    glVertex3f(m_pos.x, m_pos.y, m_pos.z);
    glVertex3f(farBottomRight.x, farBottomRight.y, farBottomRight.z);
    
    glVertex3f(m_pos.x, m_pos.y, m_pos.z);
    glVertex3f(farBottomLeft.x, farBottomLeft.y, farBottomLeft.z);
    glEnd();
    glPopAttrib();
    glPopAttrib();
}

/*! \brief Adjusts the viewing transformation according to the camera.
 *  
 *  Sets the modelview matrix according to the camera position the view point 
 *  and the up vector. If member m_frustumView is TRUE, the modelview
 *  matrix will be set up to show the camera and it's frustum.
 */
void Camera::Look()
{
    if(m_frustumView)
    {
        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity( );
        m_viewPt = m_pos + m_viewDir;

        float d = 0.2f*m_zFar;

        Vec3f pos = m_pos + d*m_up + d*m_right - d*m_viewDir;
        Vec3f lookat = m_pos + 0.4f*m_zFar*m_viewDir;

        gluLookAt(pos.x, pos.y, pos.z, lookat.x, lookat.y, lookat.z, 
                  m_up.x, m_up.y, m_up.z);

        Draw();
    }
    else
    {
        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity( );
        m_viewPt = m_pos + m_viewDir;    
        gluLookAt(m_pos.x, m_pos.y, m_pos.z, m_viewPt.x, m_viewPt.y, 
                  m_viewPt.z, m_up.x, m_up.y, m_up.z);
    }
}

/*! \brief Sets the member m_frustumView and calls Perspective().
 *
 *  This function is called every time that the global variable frustumView
 *  changes its state. FrustumView updates the member m_frustumView and calls
 *  Perspective() in order to draw/hide the cameras frustum.
 *
 *  \param frustumView Boolean which is TRUE if the frustum shall be drawn.
 */
void Camera::FrustumView(bool frustumView)
{
    m_frustumView = frustumView;
    Perspective(m_windowWidth, m_windowHeight);
}

/*! \brief Adjusts the viewing frustum according to the camera.
 *  
 *  Sets the projection matrix according to the FOV Y angle, the zoom factor,
 *  the FOV ratio and the near and far clip planes.
 *  
 *  \param width the width of the view port.
 *  \param height the height of the view port.
 */
void Camera::Perspective(GLint width, GLint height)
{
    GLfloat zFar = m_frustumView ? 2.0f*m_zFar : m_zFar;
    GLfloat zoom = m_frustumView ? 1.0f : m_zoomFactor;

    m_windowWidth = width;
    m_windowHeight = height;

    height = (height <= 0) ? 1 : height;
    glViewport( 0, 0, ( GLint )width, ( GLint )height );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    gluPerspective(zoom * m_fovY, m_fovRatio*((float)width)/((float)height), 
                    m_zNear, zFar );
    glMatrixMode( GL_MODELVIEW );
}

/*! \brief Converts an angle from degrees to radians.
 *  
 *  Converts an angle from degrees to radians.
 *  
 *  \param degrees the angle in degrees.
 *  \return The angle in radians. 
 */
GLfloat Camera::m_DegreesToRadians(GLfloat degrees)
{
    GLfloat radianAngle = (float)(degrees*M_PI/180.0);
    return radianAngle;
}

/*! \brief Moves the camera.
 *  
 *  Moves the camera according to the a specific action and a specific step.
 *  
 *  \param m the type of movement.
 *  \param amount the step.
 */
void Camera::Move(CameraMovement m, GLfloat step)
{
    switch(m)
    {
        case ELEVATE:
            m_Elevate(step);
            break;
        case FORWARD:
            m_Forward(step);
            break;
        case STRAFE:
            m_Strafe(step);
            break;
        case ZOOM:
            m_Zoom(step);
            break;
        case YAW:
            m_Yaw(step);
            break;
        case PITCH:
            m_Pitch(step);
            break;
        case ROLL:
            m_Roll(step);
            break;
    }
}

/*! \brief Update camera variables
 *
 *  Update camera variables for use in the testshell.
 *  
 *  \param pos storage for camera position.
 *  \param dir storage for view direction.
 *  \param up storage for up vector.
 */
void Camera::GetCameraVariables(GLfloat* pos, GLfloat* dir, GLfloat* up,
                                GLfloat& zoomFactor, GLfloat& fovY, 
                                GLfloat& fovRatio, GLfloat& zNear, 
                                GLfloat& zFar)
{
    pos[0] = m_pos.x;
    pos[1] = m_pos.y;
    pos[2] = m_pos.z;

    dir[0] = m_viewDir.x;
    dir[1] = m_viewDir.y;
    dir[2] = m_viewDir.z;

    up[0] = m_up.x;
    up[1] = m_up.y;
    up[2] = m_up.z;

    zoomFactor = m_zoomFactor;
    fovY = m_fovY;
    fovRatio = m_fovRatio;
    zNear = m_zNear;
    zFar = m_zFar; 
}

