/*! \file camera.h
 *  \brief Camera header file.
 *
 *  Contains the declarations of the Camera class. 
 */

#ifndef _CAMERA_H_
#define _CAMERA_H_

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef clamp
#define clamp(X, MIN, MAX) ((X<MIN) ? MIN : ((X>MAX) ? MAX : X))
#endif

#include <math.h>
#include <GL/glew.h>
#include "vec3.h"

/*! \brief The camera class.
 *         Handles movement and rotation of the camera.
 *
 *  Functionality includes elevating, moving, strafing, zooming, panning, 
 *  yawing, pitching, rolling. The camera frustum can be rendered.
 */
class Camera
{
public:
	Camera();
	~Camera();
	
    enum CameraMovement {ELEVATE, FORWARD, STRAFE, ZOOM, YAW, PITCH, 
                        ROLL}; /*!< Constants used for specifying camera 
                                    movements */

    void SetState(GLfloat x, GLfloat y, GLfloat z, GLfloat yaw, 
                  GLfloat pitch, GLfloat roll, GLfloat fovY, 
                  GLfloat fovRatio, GLfloat zNear, GLfloat zFar, 
                  GLfloat zoomFactor);
    void SetRange(GLfloat zNear, GLfloat zFar);
    void Move(CameraMovement m, GLfloat step);
	void Draw();
	void Look();
    void Perspective(GLint width, GLint height);
    void FrustumView(bool frustumView);
    void GetCameraVariables(GLfloat* pos, GLfloat* dir, GLfloat* up, GLfloat& zoomFactor, GLfloat& fovY, GLfloat& fovRatio, GLfloat& zNear, GLfloat& zFar);

private:
	Vec3f m_pos;            /*!< The position of the camera */
    Vec3f m_up;             /*!< The up direction */
    Vec3f m_viewDir;        /*!< The view direction */    
    Vec3f m_viewPt;         /*!< The viewpoint */	
    Vec3f m_right;          /*!< The right direction */
    GLfloat m_yaw;          /*!< The yaw angle */
    GLfloat m_pitch;        /*!< The pitch angle */
    GLfloat m_roll;         /*!< The roll angle */
    GLfloat m_fovY;         /*!< The field-of-view angle in the y direction */
    GLfloat m_fovX;         /*!< The field-of-view angle in the x direction */
    GLfloat m_fovRatio;     /*!< The aspect ratio of the field of view (x/y) */
    GLfloat m_zNear;        /*!< The distance from the camera to the near 
                                 clipping plane */
    GLfloat m_zFar;         /*!< The distance from the camera to the far 
                                 clipping plane */
    GLfloat m_zoomFactor;   /*!< The amount of zooming */

    bool m_frustumView;     /*!< Whether the camera shall view its frustum. */
    int m_windowWidth;      /*!< Glut window width */
    int m_windowHeight;     /*!< Glut window height */

    GLfloat m_DegreesToRadians(GLfloat degrees);

	void m_Elevate(GLfloat step);
	void m_Forward(GLfloat step);
	void m_Strafe(GLfloat step);

	void m_Zoom(GLfloat zoomFactor);
	
	void m_Yaw(GLfloat angle);
	void m_Pitch(GLfloat angle);
	void m_Roll(GLfloat angle);
};

#endif
