/*! \file display.cpp
 *  \brief Functions for rendering
 *
 */

#include "display.h"

/*! \brief Prints a string on screen
 *
 *  The function use a Bitmap font to render a string.
 *
 *  \param x the upper left x position
 *  \param y the upper left y position
 *  \param text the string to be printed
 */
void PrintString(float x, float y, const char* text)
{
    glRasterPos2d(x, y);
    while(*text != 0)
        glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *text++);
}

/*! \brief Draws on-screen info
 *
 *  Prints values and parameters on-screen.
 *
 *  \sa PrintString()
 */
void DrawOnScreenInfo()
{
    static char buffer[256];

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(-1.5, 1.5, -1, 1);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glColor3f(1.0f, 1.0f, 1.0f);

    float vpos = 0.40f;
    float hpos = -1.3f;
    float hposButton = 0.1f;

    if(keyboardFPS)
    {
        // Frames per second
        sprintf(buffer, "FPS: %d", (int)(1.0f/dt));
        PrintString(-1.4, 0.9, buffer);
   
        sprintf(buffer, "Time: %.1f", (float)t);
        PrintString(-1.4, 0.85, buffer);

        sprintf(buffer, "Altitude: %d m", (int)cameraPosition[1]);
        PrintString(-1.4, 0.8, buffer);
    }

    if(keyboardHelp)
    {
        // Generate background for help menu
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, helpTexture.texID);

        glBegin(GL_POLYGON);
        glTexCoord2f(0.0,1.0);
        glVertex2f(-1.4,-0.9);
        glTexCoord2f(1.0,1.0);
        glVertex2f(1.4,-0.9);
        glTexCoord2f(1.0,0.0);
        glVertex2f(1.4,0.75);
        glTexCoord2f(0.0,0.0);
        glVertex2f(-1.4,0.75);
        glEnd();

        glEnable(GL_CULL_FACE);
        glDisable(GL_BLEND);
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glDisable(GL_TEXTURE_2D);

        // Information
        sprintf(buffer, "PARAMETERS:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "CHANGE WITH:");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        // Sun
        sprintf(buffer, "Sun direction: [%.2f, %.2f, %.2f]",
                sunDirection[0], sunDirection[1], sunDirection[2]);
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "Q/E");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Sun color: [%.1f, %.1f, %.1f]",
                sunColor[0], sunColor[1], sunColor[2]);
        PrintString(hpos, vpos, buffer);
        vpos -= 0.05;

        //Lod
        sprintf(buffer, "Mesh quality level: %d", lodRes);
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "Z/X");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Visualization quality level: %d", lodVis);
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "C/V");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        // Water
        sprintf(buffer, "Water color: [%.1f, %.1f, %.1f]", 
                waterColor[0], waterColor[1], waterColor[2]);
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "U/J, I/K, O/L");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        Vec3f wind(windVelocity[0], 0, windVelocity[1]);
        
        sprintf(buffer, "Wind speed: %.1f", wind.Norm());
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "W/S");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        wind.Normalize();
        sprintf(buffer, "Wind direction: [%.2f, %.2f]", wind.x, wind.z);
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "A/D");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.1;

        // Help text
        sprintf(buffer, "CONTROLS (use keypad for movements).");
        PrintString(hpos, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Strafe/Move:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "Arrows");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Elevate:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "/, *");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Roll:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "HOME, PGUP");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Yaw:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "END, PGDN");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Pitch:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "INS, DEL");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Zoom:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "+, -");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Look at frustum:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "RETURN");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide FPS and time:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F2");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide wind direction:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F3");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide light direction:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F4");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide coordinate axis:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F5");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide water:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F6");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide terrain:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F7");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show/hide sky:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F8");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Toggle between wireframe/solid mode:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F9");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;

        sprintf(buffer, "Show hide floating balloon:");
        PrintString(hpos, vpos, buffer);
        sprintf(buffer, "F10");
        PrintString(hposButton, vpos, buffer);
        vpos -= 0.05;
    }
  
    glEnable(GL_DEPTH_TEST);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
}

/*! \brief Draw a 3D arrow
 *
 *  \param pos position of arrow
 *  \param dir direction of arrow
 *  \param r red
 *  \param g green
 *  \param b blue
 *  \sa DrawInfo()
 */
void DrawArrow(Vec3f pos, Vec3f dir, float r, float g, float b,
               bool inverse=false)
{
    // Set new range
    float d_near = zNear;
    float d_far = zFar;
    camera.SetRange(0.1f, 100.0f);

    GLfloat ambient[] = {r*0.4, g*0.4, b*0.4, 1.0f};
	GLfloat diffuse[] = {r*0.6, g*0.6, b*0.6, 1.0f};
    GLfloat reset_ambient[] = {0.2, 0.2, 0.2, 1};
	GLfloat reset_diffuse[] = {0.8, 0.8, 0.8, 1};

    float length = dir.Norm();
    dir.Normalize();

    Vec3f Z = Vec3f(0.0f, 0.0f, 1.0f);
    Vec3f V = Z.Cross(dir);
    float a = acosf(dir.Dot(Z))*180/3.1415f;

    GLUquadric* qobj;
    qobj = gluNewQuadric();
    gluQuadricDrawStyle(qobj, GLU_FILL);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);

    glPushMatrix();
    glTranslatef(pos.x, pos.y, pos.z);
    glRotatef(a, V.x, V.y, V.z);

    if(inverse)
        glTranslatef(0.0f, 0.0f, -length);
    gluCylinder(qobj, 0.1f, 0.1f, length, 16, 2);
    glRotatef(180, 0, 1, 0);
    gluDisk(qobj,0.0f, 0.1f, 16, 16);
    glRotatef(-180, 0, 1, 0);

    glTranslatef(0, 0, length-0.1);
    gluCylinder(qobj, 0.2f, 0.0f, 0.2f, 16, 2);
    glRotatef(180, 0, 1, 0);
    gluDisk(qobj,0.0f, 0.2f, 16, 16);
    glRotatef(-180, 0, 1, 0);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, reset_ambient);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, reset_diffuse);

    gluDeleteQuadric(qobj);
    glPopMatrix();

    // Restore range
    camera.SetRange(d_near, d_far);
}

/*! \brief Draw info in world
 *
 *  Draws axis, wind direction and light direction.
 *
 *  \sa DrawArrow()
 */
void DrawInfo()
{
    glEnable(GL_LIGHTING);
    if(keyboardAxis)
    {
        // Draw Axis
        DrawArrow(Vec3f(0.0f,0.0f,0.0f), 10.0f*Vec3f(1.0f,0.0f,0.0f),
                  1.0f, 0, 0);
        DrawArrow(Vec3f(0.0f,0.0f,0.0f), 10.0f*Vec3f(0.0f,1.0f,0.0f),
                  0, 1.0f, 0);
        DrawArrow(Vec3f(0.0f,0.0f,0.0f), 10.0f*Vec3f(0.0f,0.0f,1.0f),
                  0, 0, 1.0f);
                  
        glPushMatrix();
        glTranslatef(200,0,200);
        glBegin(GL_LINES);
        glVertex4f(0.0,0.0,0.0,1.0);
        glVertex4f(0.0,10.0,0.0,1.0);
        glEnd();
        glPopMatrix();
    }

    GLfloat M[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, M);

    // Draw wind direction
    if(keyboardWind)
    {
        Vec3f pos(0.0f,0.0f,-10.0f);
        Vec3f pos1(0.0f,0.0f,-10.0f);
        Vec3f dir(windVelocity[0]*0.25f, 0.0f, windVelocity[1]*0.25f);

        pos1.x = pos1.x - M[12];
        pos1.y = pos1.y - M[13];
        pos1.z = pos1.z - M[14];

        pos.x = pos1.x * M[0] + pos1.y * M[1] + pos1.z * M[2];
        pos.y = pos1.x * M[4] + pos1.y * M[5] + pos1.z * M[6];
        pos.z = pos1.x * M[8] + pos1.y * M[9] + pos1.z * M[10];

        DrawArrow(pos, dir, 1.0f, 0, 1.0f);
    }

    // Draw light direction
    if(keyboardLight)
    {
        Vec3f dir(sunDirection[0], sunDirection[1], sunDirection[2]);
        Vec3f pos = Vec3f(0.0f,0.0f,-10.0f);
        Vec3f pos1 = pos;

        pos1.x = pos1.x - M[12];
        pos1.y = pos1.y - M[13];
        pos1.z = pos1.z - M[14];

        pos.x = pos1.x * M[0] + pos1.y * M[1] + pos1.z * M[2];
        pos.y = pos1.x * M[4] + pos1.y * M[5] + pos1.z * M[6];
        pos.z = pos1.x * M[8] + pos1.y * M[9] + pos1.z * M[10];

        DrawArrow(pos, dir, 1.0f, 1.0f, 0.0f, true);
    }
}

/*! \brief Draws a simple terrain
 *
 */
void DrawTerrain()
{
    glColor3f(1.0, 1.0, 1.0);

    groundShader.Activate();

    // Setup textures
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, groundTexture.texID);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_2D, groundNoiseTexture.texID);

    groundShader.UpdateUniforms();

    groundMesh.DrawVBO();

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_2D);

    groundShader.Deactivate();
}

/*! \brief Draws a sky
 *
 */
void DrawSky()
{
    // cameraPosition is set in input.cpp
    float h0 = 1000.0f+1.0f*cameraPosition[1]; // Height over camera.
    float ratio = 0.5f; // If ratio is changed, update UpdateInput()!
    skyHeight = h0+cameraPosition[1];
    float r = skyHeight / ratio; // Radius of sphere
    skyOffset = skyHeight - r;
    waterHorizon = sqrt(r*r-skyOffset*skyOffset);   //Distance to the horizon

    glColor3f(1.0,1.0,1.0);

    skyShader.Activate();

    skyShader.UpdateUniforms();

    // Setup textures
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glBindTexture(GL_TEXTURE_2D, skyTexture.texID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glActiveTextureARB(GL_TEXTURE1_ARB);
    glBindTexture(GL_TEXTURE_2D, skyCloudTexture.texID);

    glCullFace(GL_FRONT);
    glPushMatrix();
    glTranslatef(cameraPosition[0], skyOffset, cameraPosition[2]);
    glutSolidSphere(r, 32, 32);
    glPopMatrix();
    glCullFace(GL_BACK);

    skyShader.Deactivate();
}

/*! \brief Draws balloon floating on the water.
 *
 */
void DrawBalloon()
{
    float pos[2] = {150, 150};
    glColor3f(1.0,0.0,0.0);

    glPushMatrix();
    glTranslatef(pos[0], OV::OVGetHeight(pos[0], pos[1]), pos[1]);
//    glTranslatef(0.0, 100.0, 0.0);
    glutSolidSphere(2.0, 16, 16);
    glPopMatrix();
}

/*! \brief Draws a boat floating on the water
 *
 */
void DrawBoat()
{
    float l = 10.0f;
    float w = l/10.0f;
    float h = l/4.0f;
    float c = l/10.0f;

    static Vec3f stern(150.0f, 0.0f, 150.0f);
    static Vec3f bow = stern + Vec3f(l, 0.0f, 0.0f);
    static Vec3f port(stern.x+l/2.0f, 0.0f, stern.z-w);
    static Vec3f starboard(stern.x+l/2.0f, 0.0f, stern.z+w);

    float alpha=0.7f;
    stern.y = (alpha*stern.y + 
              (1.0f-alpha)*OV::OVGetHeight(stern.x, stern.z));
    bow.y = (alpha*bow.y + (1.0f-alpha)*OV::OVGetHeight(bow.x, bow.z));
    port.y = (alpha*port.y + (1.0f-alpha)*OV::OVGetHeight(port.x, port.z));
    starboard.y = (alpha*starboard.y + 
                  (1.0f-alpha)*OV::OVGetHeight(starboard.x, starboard.z));

    Vec3f dir = bow - stern;
    dir.Normalize();
    float a = acosf(dir.Dot(Vec3f(1.0f, 0.0f, 0.0f)));
    a *= 180.0f;
    a /= 3.141592f;

    dir = starboard-port;
    dir.Normalize();
    float b = acosf(dir.Dot(Vec3f(0.0f, 0.0f, 1.0f)));
    b *= 180.0f;
    b /= 3.141592f;

    glPushMatrix();
    glTranslatef(stern.x, stern.y-0.25f*h, stern.z);
    if(bow.y > stern.y)
        glRotatef(a, 0, 0, 1);
    else
        glRotatef(-a, 0, 0, 1);

    if(starboard.y > port.y)
        glRotatef(-b, 1, 0, 0);
    else
        glRotatef(b, 1, 0, 0);

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, boatSideTexture.texID);
    
    // port
    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glVertex3f(0.0f, 0.0f, -w);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(l, 0.0f, -w);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(l, h, -w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(0.0f, h, -w);
    glEnd();

    // starboard
    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glVertex3f(0.0f, 0.0f, w);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(l, 0.0f, w);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(l, h, w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(0.0f, h, w);
    glEnd();

    // stern
    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glVertex3f(0.0f, 0.0f, -w);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(0.0f, 0.0f, w);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(0.0f, h, w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(0.0f, h, -w);
    glEnd();

    // bow
    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(l, 0.0f, -w);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(l, h, -w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(l+h, h, 0.0f);
    glEnd();

    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(l, 0.0f, w);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(l, h, w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(l+h, h, 0.0f);
    glEnd();

    glBegin(GL_POLYGON);
    glTexCoord2f(1.0f,1.0f);
    glVertex3f(l, 0.0f, -w);
    glTexCoord2f(0.0f,1.0f);
    glVertex3f(l, 0.0f, w);
    glTexCoord2f(0.0f,0.0f);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(l+h, h, 0.0f);
    glEnd();

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);

    // Deck
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, boatDeckTexture.texID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(0.0f, h, -w);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(0.0f, h, w);
    glTexCoord2f(1.0f,5.0f);
    glVertex3f(l, h, w);
    glTexCoord2f(0.0f,5.0f);
    glVertex3f(l, h, -w);
    glEnd();
    
    glBegin(GL_POLYGON);
    glTexCoord2f(0.0f,0.0f);
    glVertex3f(l, h, -w);
    glTexCoord2f(0.0f,1.0f);
    glTexCoord2f(1.0f,1.0f);
    glVertex3f(l+h, h, 0.0f);
    glTexCoord2f(1.0f,0.0f);
    glVertex3f(l, h, w);
    glEnd();

    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);

    // Top
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, boatTopTexture.texID);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glTranslatef(l/5.0f,h+c/4.0f,0.0f);
    glScalef(0.75f,0.5f,1.0f);
    glutSolidCube(c);
    glScalef(1.25f,2.0f,1.0f);
    glTranslatef(0.0f,c*0.75f,0.0f);
    glScalef(0.75f,0.5f,1.0f);
    glutSolidCube(2.0f*c);

    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_2D);

    glPopMatrix();
}

/*! \brief Draw scene
 *
 *  This function calls all the other draw methods to generate
 *  the scene.
 *
 *  \sa DrawTerrain(), OVDrawWater(), DrawInfo() and DrawOnScreenInfo()
 */
void DrawScene()
{
    //Floating balloon
    if(keyboardDrawBoat)
        DrawBoat();
    // Water
    if(keyboardDrawWater) {
        OV::OVDrawWater();
        if(keyboardLook)
            OV::OVDrawProjectorFrustum();
    }

    // Terrain
    if(keyboardDrawTerrain)
        DrawTerrain();

    // Sky
    if(keyboardDrawSky)
        DrawSky();



    // Info
    DrawInfo();

    // On-screen info text
    DrawOnScreenInfo();
}
