/*! \file input.cpp
 *  \brief Input source file.
 *
 *  Contains the source code for keyboard inputs. Whenever a button used to
 *  control the camera movements is pressed, the code tells the camera to
 *  to update it's position and zooming etc. Especially, if 'F1' or 'escape'
 *  is pressed the program shows the help menu or quit the application,
 *  respectively.
 */

#include "input.h"

/*! \brief Brief description of KeyDownFunc.
 *
 *  This function triggers whenever a "non-special-button" is pressed, which
 *  for this application means any button on the num pad or the escape button.
 *  Whenever a button that controls the camera i pressed a flag will be set.
 *  This flag is used by the function UpdateInput() to let the camera know in
 *  which direction it should move.
 *
 *  \param key The actual key that was pressed.
 *  \param x The x coordinate for the mouse pointer (not needed here).
 *  \param y The y coordinate for the mouse pointer (not needed here).
 */
void KeyDownFunc(unsigned char key, int x, int y)
{
    // Speed boost
    int mod = glutGetModifiers();
    if (mod == (GLUT_ACTIVE_SHIFT))
      moveSpeed = 400.0f;
    else 
      moveSpeed = 40.0f;

    switch(key)
    {
    case '1': // Yaw left
        keyboardYaw = 1;
        break;
    case '3': // Yaw right
        keyboardYaw = -1;
        break;
    case '7': //Roll left
        keyboardRoll = 1;
        break;
    case '9': // Roll right
        keyboardRoll = -1;
        break;
    case ',': // Pitch up
        keyboardPitch = 1;
        break;
    case '0': // Pitch down
        keyboardPitch = -1;
        break;
    case '*': // Elevate
        keyboardElevate = 1;
        break;
    case '/': // Lower
        keyboardElevate = -1;
        break;
    case '+': // Zoom in
        keyboardZoom = -1;
        break;
    case '-': // Zoom out
        keyboardZoom = 1;
        break;
    case '8': // Move forward
        keyboardForward = 1;
        break;
    case '2': // Move backwards
        keyboardForward = -1;
        break;
    case '6': // Strafe right
        keyboardStrafe = 1;
        break;
    case '4': // Strafe left
        keyboardStrafe = -1;
        break;
    case 13: // Frustum view (Carriage Return)
       {
            keyboardLook = !keyboardLook;
            camera.FrustumView(keyboardLook);
       }
       break;
    case 'e': case 'E': // Sun direction
        keyboardMoveSun = -1;
        break;

    case 'q': case 'Q': // Sun direction
        keyboardMoveSun = 1;
        break;
    case 'a': case 'A': // Wind direction
        keyboardTurnWind = 1;
        break;
    case 'd': case 'D': // Wind direction
        keyboardTurnWind = -1;
        break;
    case 'w': case 'W': // Wind speed
        keyboardMoveWind = 1;
        break;
    case 's': case 'S': // Wind speed
        keyboardMoveWind = -1;
        break;
    case 'z': case 'Z': // LOD resolution
        {
            if(lodRes > 1)
            {
                lodRes--;
                OV::OVSetResolutionLOD(lodRes);
            }
        }
        break;
    case 'x': case 'X': // LOD resolution
        {
            if(lodRes < 3)
            {
                lodRes++;
                OV::OVSetResolutionLOD(lodRes);
            }
        }
        break;
    case 'c': case 'C': // LOD visualization
        {
            if(lodVis > 1)
            {
                lodVis -= 1;
                OV::OVSetVisualizationLOD(lodVis);
            }
        }
        break;
    case 'v': case 'V': // LOD visualization
        {
            if(lodVis < 3)
            {
                lodVis += 1;
                OV::OVSetVisualizationLOD(lodVis);
            }
        }
        break;
    case 'n': case 'N': // Wave amplitude
        {
            waveAmpFactor += 0.05;
            OV::OVSetWaveAmpFactor(waveAmpFactor);
        }
        break;
    case 'm': case 'M':
        {
            waveAmpFactor -= 0.05;
            OV::OVSetWaveAmpFactor(waveAmpFactor);
        }
        break;
    case 'u': case 'U': // Water color - Red
        waterColor[0] = clamp(waterColor[0]+0.1f, 0.0f, 1.0f);
        break;
    case 'j': case 'J': // Water color - Red
        waterColor[0] = clamp(waterColor[0]-0.1f, 0.0f, 1.0f);
        break;
    case 'i': case 'I': // Water color - Green
        waterColor[1] = clamp(waterColor[1]+0.1f, 0.0f, 1.0f);
        break;
    case 'k': case 'K': // Water color - Green
        waterColor[1] = clamp(waterColor[1]-0.1f, 0.0f, 1.0f);
        break;
    case 'o': case 'O': // Water color - Blue
        waterColor[2] = clamp(waterColor[2]+0.1f, 0.0f, 1.0f);
        break;
    case 'l': case 'L': // Water color - Blue
        waterColor[2] = clamp(waterColor[2]-0.1f, 0.0f, 1.0f);
        break;
    case 27: // Quit (Escape)
        exit(0);
        break;
    default:
        break;
    }
}

/*! \brief Brief description of SpecDownFunc.
 *
 *  This function triggers whenever a "special" button is pressed.
 *  For this application, the only special button is 'F1'. When this button
 *  is pressed a help menu appears on the screen. If the button is pressed
 *  once again, the help menu disappears.
 *
 *  \param key The actual key that was pressed.
 *  \param x The x coordinate for the mouse pointer (not needed here).
 *  \param y The y coordinate for the mouse pointer (not needed here).
 */
void SpecDownFunc(int key, int x, int y)
{
    switch(key)
    {
    case GLUT_KEY_F1: // Show help menu
        keyboardHelp = !keyboardHelp;
        break;
    case GLUT_KEY_F2: // Toggle display of FPS and time
        keyboardFPS = !keyboardFPS;
        break;
    case GLUT_KEY_F3: // Toggle wind arrow
        keyboardWind = !keyboardWind;
        break;
    case GLUT_KEY_F4: // Toggle light arrow
        keyboardLight = !keyboardLight;
        break;
    case GLUT_KEY_F5: // Toggle axis
        keyboardAxis = !keyboardAxis;
        break;

    case GLUT_KEY_F6: // Toggle water
        keyboardDrawWater = !keyboardDrawWater;
        break;
    case GLUT_KEY_F7: // Toggle terrain
        keyboardDrawTerrain = !keyboardDrawTerrain;
        break;
    case GLUT_KEY_F8: // Toggle sky
        keyboardDrawSky = !keyboardDrawSky;
        break;
    case GLUT_KEY_F9: // Toggle wireframe
        {
            if(polygonMode == GL_FILL)
                polygonMode = GL_LINE;
            else
                polygonMode = GL_FILL;
        }
        break;

    case GLUT_KEY_F10: // Toggle floating boat
        keyboardDrawBoat = !keyboardDrawBoat;
        break;

    case GLUT_KEY_F11:
        {
            static int mode = 0;

            switch(mode)
            {
            case 0:
                camera.SetState(0, 300, -700, 0, -18, 0, 45, 1, zNear, zFar, 1);
                sunAngle = 3.1415927/2.0f;
                windSpeed = 10.0f;
                windDir = 0.0f;
                keyboardDrawBoat = false;
                break;
            case 1:
                camera.SetState(0, 20, 0, 90, 0, 0, 45, 1, zNear, zFar, 1);
                sunAngle = 0.1f;
                windSpeed = 5.0f;
                windDir = -1.5f;
                keyboardDrawBoat = false;
                break;
            case 2:
                camera.SetState(200, 10, 202, 145, 0, 0, 45, 1, zNear, zFar, 1);
                sunAngle = 3.1415927/2.0f;
                windSpeed = 15.0f;
                windDir = 0.54f;
                keyboardDrawBoat = true;
                break;
            }

            mode = (mode + 1) % 3;

        }
        break;

    default:
        break;
    }
}

/*! \brief Brief description of KeyUpFunc.
 *
 *  This function triggers whenever a "non-special-button" is released.
 *  Whenever a button on the num pad is released, the function clears
 *  that buttons flag. This flag is used by the function UpdateInput()
 *  to let the camera know when to stop it's movements etc.
 *
 *  \param key The actual key that was pressed.
 *  \param x The x coordinate for the mouse pointer (not needed here).
 *  \param y The y coordinate for the mouse pointer (not needed here).
 *  \return void
 */
void KeyUpFunc(unsigned char key, int x, int y)
{
    switch(key)
    {
    case '1': case '3':// Yaw
        keyboardYaw = 0;
        break;
    case '7': case '9': // Roll
        keyboardRoll = 0;
        break;
    case ',': case '0': // Pitch
        keyboardPitch = 0;
        break;
    case '*': case '/': // Elevate
        keyboardElevate = 0;
        break;
    case '+': case '-': // Zoom
        keyboardZoom = 0;
        break;
    case '8': case '2': // Move
        keyboardForward = 0;
        break;
    case '6': case '4': // Strafe
        keyboardStrafe = 0;
        break;

    case 'e': case 'E': case 'q': case 'Q': // Move sun
        keyboardMoveSun = 0;
        break;

    case 's': case 'S': case 'w': case 'W': // Change wind speed
        keyboardMoveWind = 0;
        break;

    case 'a': case 'A': case 'd': case 'D': // Move wind dir
        keyboardTurnWind = 0;
        break;

    default:
        break;
    }
}

/*! \brief Brief description of SpecUpFunc.
 *
 *  This function triggers whenever a "special" button is released.
 *
 *  \param key The actual key that was pressed.
 *  \param x The x coordinate for the mouse pointer (not needed here).
 *  \param y The y coordinate for the mouse pointer (not needed here).
 */
void SpecUpFunc(int key, int x, int y)
{
}

/*! \brief Brief description of UpdateInput.
 *
 *  This is the function that tells the camera in which way to move.
 *  It uses global flags to determine which buttons that has been pressed
 *	by the user. It handles movements such as forward, backward, strafe,
 *  zoom, yaw, pitch, roll and a special move to look at the cameras
 *  frustum.
 */
void UpdateInput()
{
    camera.Move(camera.ELEVATE, keyboardElevate*moveSpeed*dt);
    camera.Move(camera.FORWARD, keyboardForward*moveSpeed*dt);
    camera.Move(camera.STRAFE, keyboardStrafe*moveSpeed*dt);

    camera.Move(camera.YAW, keyboardYaw*turnSpeed*dt);
    camera.Move(camera.PITCH, keyboardPitch*turnSpeed*dt);
    camera.Move(camera.ROLL, keyboardRoll*turnSpeed*dt);

    if(keyboardZoom != 0)
    {
        camera.Move(camera.ZOOM, keyboardZoom*zoomSpeed*dt);
        camera.Perspective(windowWidth, windowHeight);
    }

    camera.GetCameraVariables(cameraPosition, cameraDirection, cameraUpVector,
                              zoomFactor, fovY, fovRatio, zNear, zFar);

    // Calculate near and far clip plane
    float h = cameraPosition[1];
    float h0 = 1000.0f+1.0f*h;
    float h1 = h0 + h;
    float ratio = 0.5f;
    zFar = sqrtf(4*h1*h1-2*h0*h1+h0*h0)*1.1f;
    zNear = zFar / 1000.0f;
    camera.SetRange(zNear, zFar);

    // Update variables for horizon
    skyHeight = h0+cameraPosition[1];
    float r = skyHeight / ratio; // Radius of sphere
    skyOffset = skyHeight - r;
    waterHorizon = sqrt(r*r-skyOffset*skyOffset);   //Distance to the horizon

    // Move sun
    sunAngle += sunSpeed*DEG2RAD*keyboardMoveSun*dt;
    sunAngle = clamp(sunAngle, 0.1f, 3.1415927f-0.1f);

    daytime = sunAngle / M_PI / 2;

    sunDirection[0] = cosf(sunAngle);
    sunDirection[1] = -sinf(sunAngle)*1/sqrtf(2);
    sunDirection[2] = -sinf(sunAngle)*1/sqrtf(2);

    sunPosition[0] = -sunDirection[0]*10000.0f;
    sunPosition[1] = -sunDirection[1]*10000.0f;
    sunPosition[2] = -sunDirection[2]*10000.0f;

    sunColor[0] = 0.9f;
    sunColor[1] = clamp(sinf(sunAngle)+0.6, 0.0,0.9);
    sunColor[2] = clamp(sinf(sunAngle)+0.3, 0.0,0.5);

    // Update wind
    windSpeed += windMoveSpeed*keyboardMoveWind*dt;
    windSpeed = clamp(windSpeed, 0.1f, 25.0f);
    windDir += DEG2RAD*turnSpeed*keyboardTurnWind*dt;
    windVelocity[0] = windSpeed*cosf(windDir);
    windVelocity[1] = windSpeed*sinf(windDir);
}
