/*! \file mesh.cpp
 *  \brief Mesh source file.
 *
 */

#include "mesh.h"

#define odd(x) (x%2)

/*! \brief Constructor for mesh.
 *
 */
Mesh::Mesh() 
{
    m_numVertices = 0;
    m_numFaces = 0;
    hmin=hmax=0.0f;
}

/*! \brief Destructor for mesh.
 *
 */
Mesh::~Mesh()
{
    m_Clear();
}

/*! \brief Generates a grid.
 *
 *  Generates a rectangular grid between (x0,y0) and (x1,y1). If vertical is
 *  TRUE the grid will be in the xy-plane, otherwise it will be in the
 *  xz-plane.
 *
 *  \param resX the resolution in the x direction.
 *  \param resY the resolution in the y(z) direction.
 *  \param x0 the x component of the starting point.
 *  \param x1 the x component of the ending point.
 *  \param y0 the y component of the starting point.
 *  \param y1 the y component of the ending point.
 *  \param vertical decides if the grid is vertical or horizontal.
 *  \sa m_AddVertex()
 */
void Mesh::GenerateGrid(int resX, int resY, float x0, float x1, float y0,
						float y1, bool vertical)
{
    // Clear already loaded mesh
    m_Clear();

    // Genrerate vertex list
    float dx = (x1-x0) / (resX-1);
    float dy = (y1-y0) / (resY-1);

    m_vertArray = new GLfloat[resX*resY*3*2];
    m_numVertices = 0;

    Vec3f normal_vert(0,0,1);
    Vec3f normal_hori(0,1,0);
  
    for(int y=0; y<resY; y++)
    {
      for(int x=0; x<resX; x++)
      {
        if(vertical)
          m_AddVertex(x0+x*dx, y0+y*dy, 0.0f, normal_vert);
        else
          m_AddVertex(x0+x*dx, 0.0f, y0+y*dy, normal_hori);
      }
    }
  
    // Generate index list
    m_faceArray = new GLuint[2*(resY-1)*resX];
      
    int i=0;
  
    for(int y=0; y<resY-1; y++)
    {
      if(odd(y))
      {
        for(int x=resX-1; x>=0; x--)
        {
          m_faceArray[i++] = (resX*(y+1)+x);
          m_faceArray[i++] = (resX*y+x);
        }
      }
      else
      {
        for(int x=0; x<resX; x++)
        {
          m_faceArray[i++] = (resX*y+x);
          m_faceArray[i++] = (resX*(y+1)+x);
        }
      }
    }
    m_numFaces = i-2;
}

/*! \brief Generate grid from heightmap texture
 *
 *  Load a TGA-image and create a mesh with the same resolution and set
 *  the gray-scale values as height.
 *
 *  \param filename filename
 *  \param x0 x lower bound
 *  \param x1 x upper bound
 *  \param z0 z lower bound
 *  \param z1 z upper bound
 *  \param h0 y lower bound
 *  \param h1 y upper bound
 *  \return TRUE if mesh is successfully loaded
 */
bool Mesh::LoadHeightMap(const char* filename, float x0, float x1, float z0,
                         float z1, float h0, float h1)
{
    hmin=h0;
    hmax=h1;

    // Clear already loaded mesh
    m_Clear();

    int width, height, bpp;
    char* data;

    if(!Tga2char(filename, width, height, bpp, &data))
    {
        std::cerr<<"Could not load heightmap file."<<std::endl;
        return false;
    }

    // Genrerate vertex list
    float dx = (x1-x0) / (width-1);
    float dz = (z1-z0) / (height-1);

    m_vertArray = new GLfloat[width*height*3*2];
    m_numVertices = 0;

    for(int z=0; z<height; z++)
    {
        for(int x=0; x<width; x++)
        {
            float h;
            Vec3f normal;
            m_ReadHeight(x, z, dx, dz, h0, h1, (unsigned char*)data, bpp/8,
                         width, height, h, normal);
            m_AddVertex(x0+x*dx, h, z0+z*dz, normal);
        }
    }

    // Generate index list
    m_faceArray = new GLuint[2*(height-1)*width];
   
    int i=0;
  
    for(int z=0; z<height-1; z++)
    {
        if(odd(z))
        {
            for(int x=width-1; x>=0; x--)
            {
                m_faceArray[i++] = (width*(z+1)+x);
                m_faceArray[i++] = (width*z+x);
            }
        }
        else
        {
            for(int x=0; x<width; x++)
            {
                m_faceArray[i++] = (width*z+x);
                m_faceArray[i++] = (width*(z+1)+x);
            }
        }
    }
    m_numFaces = i-2;

    free(data);
    
    return true;
}

/*! \brief Creates a Vertex Buffer Object (VBO)
 *
 */
void Mesh::CreateVBO()
{
    // Vertices
    glGenBuffersARB(1, &m_vertexVBO);
  	glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVBO);
  	glBufferDataARB(GL_ARRAY_BUFFER_ARB, m_numVertices*3*2*sizeof(GLfloat),
                    m_vertArray, GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

    // Indices
    glGenBuffersARB(1, &m_indexVBO);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexVBO);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, (m_numFaces+2)*
                    sizeof(GLuint), m_faceArray, GL_STATIC_DRAW_ARB);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
}

/*! \brief Deletes a Vertex Buffer Object (VBO)
 *
 */
void Mesh::DeleteVBO()
{
    glDeleteBuffersARB(1, &m_vertexVBO);
  	glDeleteBuffersARB(1, &m_indexVBO);
}

/*! \brief Draws the grid in OpenGL.
 *
 */
void Mesh::DrawOpenGL()
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glVertexPointer(3, GL_FLOAT, 6*sizeof(GLfloat), m_vertArray);
    glNormalPointer(GL_FLOAT, 6*sizeof(GLfloat), m_vertArray+3);
    glDrawElements(GL_TRIANGLE_STRIP, m_numFaces+2, GL_UNSIGNED_INT, m_faceArray);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
}

/*! \brief Draws the grid as a Vertex Buffer Object (VBO)
 *
 */
void Mesh::DrawVBO()
{
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, m_vertexVBO);
  	glVertexPointer(3, GL_FLOAT, 6*sizeof(GLfloat), 0);
    glNormalPointer(GL_FLOAT, 6*sizeof(GLfloat), (GLvoid*)(3*sizeof(GLfloat)));
  	glEnableClientState(GL_VERTEX_ARRAY);
  	glEnableClientState(GL_NORMAL_ARRAY);

  	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m_indexVBO);
	
  	glDrawElements(GL_TRIANGLE_STRIP, m_numFaces+2, GL_UNSIGNED_INT, 0);
  
    glDisableClientState(GL_VERTEX_ARRAY);
  	glDisableClientState(GL_NORMAL_ARRAY);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
}

/*! \brief Calculates the number of triangles (faces) in the grid.
 *
 *  \return the number of triangles.
 */
unsigned int Mesh::GetNumTriangles()
{
    return m_numFaces;
}

/*! \brief Convert pixel values to a height value and a normal.
 *
 *  \param x x position
 *  \param z z position
 *  \param dx x scale
 *  \param dz z scale
 *  \param h0 lower bound for mesh
 *  \param h1 upper bound for mesh
 *  \param data pointer to pixel data
 *  \param bytesPerPixel format of pixel data
 *  \param width width of height texture
 *  \param height height of height texture
 *  \param h height at given point
 *  \param n normal at given point
 */
void Mesh::m_ReadHeight(int x, int z, float dx, float dz, float h0, float h1,
                        unsigned char* data, int bytesPerPixel,
                        int width, int height, float& h, Vec3f& n)
{
    h = h0 + (h1-h0)*data[(width*z+x)*bytesPerPixel+bytesPerPixel-1] / 255.0f;

    float hx1=0.0f, hx2=0.0f, hz1=0.0f, hz2=0.0f;

    hx1 = (x>0) ? h0+(h1-h0)*data[(width*z+x-1)*bytesPerPixel+
                                  bytesPerPixel-1] / 255.0f:
                  h0+(h1-h0)*data[(width*z+x)*bytesPerPixel+
                                  bytesPerPixel-1] / 255.0f;

    hx2 = (x<width-1) ? h0+(h1-h0)*data[(width*z+x+1)*bytesPerPixel+
                                        bytesPerPixel-1] / 255.0f:
                        h0+(h1-h0)*data[(width*z+x)*bytesPerPixel+
                                        bytesPerPixel-1] / 255.0f;

    hz1 = (z>0) ? h0+(h1-h0)*data[(width*(z-1)+x)*bytesPerPixel+
                                  bytesPerPixel-1] / 255.0f:
                  h0+(h1-h0)*data[(width*z+x)*bytesPerPixel+
                                  bytesPerPixel-1] / 255.0f;

    hz2 = (z<height-1) ? h0+(h1-h0)*data[(width*(z+1)+x)*bytesPerPixel+
                                         bytesPerPixel-1] / 255.0f:
                         h0+(h1-h0)*data[(width*z+x)*bytesPerPixel+
                                         bytesPerPixel-1] / 255.0f;

    Vec3f down(0.0f, hz2-hz1, 2*dz);
    Vec3f right(2*dx, hx2-hx1, 0.0f);

    n = down.Cross(right);
    n.Normalize();
}

/*! \brief Adds a vertex in the list of vertices.
 *
 *  \param x x position
 *  \param y y position
 *  \param z z position
 *  \param normal normal vector
 */
void Mesh::m_AddVertex(float x, float y, float z, Vec3f& normal)
{
    int index = m_numVertices*6;
    m_vertArray[index] = x;
    m_vertArray[index+1] = y;
    m_vertArray[index+2] = z;

    m_vertArray[index+3] = normal.x;
    m_vertArray[index+4] = normal.y;
    m_vertArray[index+5] = normal.z;

    m_numVertices++;
}

/*! \brief Clears the vertex list and the face list.
 *
 */
void Mesh::m_Clear()
{
    if(m_vertArray)
        delete[] m_vertArray;

    if(m_faceArray)
        delete[] m_faceArray;
}
