/*! \file texture2d.cpp
 *  \brief Texture reading source file.
 *
 *  This file handles data from a TGA image and converts the data into
 *  a texture.
 */

#include "texture2d.h"

/*! \brief The constructor of the Texture struct
 *
 *  Initializes all variables.
 */
Texture2d::Texture2d()
{
    texID = 0;
    height = 0;
    width = 0;
    filename = 0;
    type = GL_RGBA;
    bpp=0;
}

/*! \brief The constructor of the Texture struct
 *
 *  Initializes all variables and loads the given texture.
 *
 *  \param texture the name of the file which will be used as a texture.
 */
Texture2d::Texture2d(const char* texture)
{
    Load(texture);
}

/*! \brief The destructor of the Texture struct */
Texture2d::~Texture2d()
{
    if(filename != 0)
        delete[] filename;
}

/*! \brief Loads a given texture to VRAM
 *
 *  \param texture The filename of the texture
 */
void Texture2d::Load(const char* texture, bool useFloat32)
{
  	filename = new char[strlen(texture)+1];
    strcpy(filename, texture);
    char* data;
    if( Tga2char(filename, width, height, bpp, &data) )
    {
    	// Build A Texture From The Data
    	glGenTextures(1, &texID);
    	glBindTexture(GL_TEXTURE_2D, texID);
    	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

        // Is TGA 24 bits? If so, set 'type' to GL_RGB (GL_RGBA is default)
        switch(bpp)
        {
        case 8:
            type = GL_LUMINANCE;
            break;
            
        case 24:
            type = GL_RGB;
            break;
        
        case 32:
            type = GL_RGBA;
            break;
        }

        if(useFloat32)
        {
    	    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, type,
                         GL_UNSIGNED_BYTE, (GLubyte*)data); 
        }
        else
        {
    	    glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type,
                         GL_UNSIGNED_BYTE, (GLubyte*)data); 
        }

        free(data);
    }
}

/*! \brief Converts a TGA image to a string of characters
 *
 *  \param filename the filename of the TGA image
 *  \param width image width
 *  \param height image height
 *  \param bpp bits per pixel
 *  \param data storage for the texture
 *  \return TRUE if the function ended with success
 */
bool Tga2char(const char* filename, int& width, int& height, int& bpp,
              char** data)
{
    // Uncompressed Header RGB
	GLubyte		RGBheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};
    // Uncompressed header Greyscale
    GLubyte     Greyheader[12]={0,0,3,0,0,0,0,0,0,0,0,0};
    // Flag for TGA type: RGB (true) or Greyscale (false)
    bool        TGARGB=true;
    // Used To Compare TGA Header
	GLubyte		TGAcompare[12];
    // Storage for header data
	GLubyte		header[6];
    // Bytes used to store a pixel
    GLubyte     bytesPerPixel;
    // Bytes used to stora a row in image
    GLuint      bytesPerRow;
    GLubyte     origin_upper; // Origin
	GLuint		imageSize;    // Image size
    GLubyte*    imageData;    // Image data
	GLuint		temp;


    std::cout << "opening image file: "<< filename<<std::endl;

    // Open The TGA File
    FILE *file = fopen(filename, "rb");

    // Check if file exist
    if(file == NULL)
    {
        std::cerr<<"Could not find image."<<std::endl;
        return false;
    }
   
    // Check if there are 12 bytes to read and if the header
    // is a correct TGA header
    if(fread(TGAcompare,1,sizeof(TGAcompare),file)==sizeof(TGAcompare))
    {
        if(memcmp(RGBheader,TGAcompare,sizeof(RGBheader))!=0 &&
           memcmp(Greyheader,TGAcompare,sizeof(Greyheader))!=0)
        {
            fclose(file);
            std::cerr<<"Image header first 12 bytes incorrect."<<std::endl;
            return false;
        }
    }
    else
    {
        fclose(file);
        std::cerr<<"Could not read image data."<<std::endl;
        return false;
    }

    // Read in the useful last 6 bytes from image header
    if(fread(header,1,sizeof(header),file)!=sizeof(header))
    {
        fclose(file);
        std::cerr<<"Image header last 6 bytes incorrect."<<std::endl;
        return false;
    }

    // Determine image type: RGB or Greyscale
    if(TGAcompare[2] == 3)
        TGARGB = false;

    // Determine the TGA width and height (highbyte*256+lowbyte)
	width  = header[1] * 256 + header[0];
	height = header[3] * 256 + header[2];

    // Check width and height and if TGA is 24 or 32 bits    
 	if(width <=0 || height <=0)
	{
		fclose(file);
        std::cerr <<"Width and/or height is less than or equal to zero."
                  <<std::endl;
        return false;
	}
    if(header[4]!=24 && header[4]!=32 && header[4] != 8)
    {
        fclose(file);
        std::cerr<<"Bits per pixel is not 8 (greyscale)"
                 <<", 24 (RGB) or 32 (RGBA)."<<std::endl;
        return false;
    }

	bpp	= header[4];		// Grab The TGA's Bits Per Pixel (8, 24 or 32)
	bytesPerPixel = bpp/8;	// Divide By 8 To Get The Bytes Per Pixel

    origin_upper = header[5] & 0x20;

    // Calculate The Memory Required For The TGA Data
    imageSize = width*height*bytesPerPixel;

    // Reserve Memory To Hold The TGA Data
	imageData=(GLubyte *)malloc(imageSize);

    fseek(file, 18, SEEK_SET);

    if(imageData != NULL)
    {
        unsigned int bytesRead = 0;
        // Does storage memory exist and do image size match memory?
        bytesPerRow = width*bytesPerPixel;
        for(int y=0; y<height; y++)
        {
            int row = origin_upper ? y : height-y-1;
            bytesRead += (unsigned int)fread(&imageData[row*bytesPerRow], 1,
                                             bytesPerRow, file);
        }

        if(bytesRead != imageSize)
        {
            std::cerr<<"Error reading data"<<std::endl;
            free(imageData);
            return false;
        }
    }
    else
    {
		fclose(file);
        std::cerr<<"Storage memory may not exist, or image size != memory" 
                 <<std::endl;
        return false;
    }
    
    // Loop through image data and swap 1st and 3rd bytes (BGR -> RGB)
    if(TGARGB)
    {
	    for(GLuint i=0; i<imageSize; i+=bytesPerPixel)
	    {
		    temp=imageData[i];	
		    imageData[i] = imageData[i + 2];
		    imageData[i + 2] = temp;
	    }
    }

	fclose (file);

    *data = (char*)imageData;

    return true;
}


