/*! \file timer.cpp
 *  \brief Timer file.
 *
 *  A high performance timer for Windows and Linux.
 */

#include "timer.h"

/*! \brief Timer constructor
 *
 *  Reads the counter frequency if in Windows and resets time.
 */
Timer::Timer()
{
#ifdef WIN32
    QueryPerformanceFrequency(&m_freq);
#endif
    ResetTime();
}

/*! \brief Get current time.
 * 
 *  Reads the QueryPerformanceCounter if in Windows or uses
 *  gettimeofday to read current time. Then converts it to seconds.
 *
 *  \return Current time in seconds.
 */
double Timer::GetTime()
{
#ifndef WIN32
    timeval tim;
    gettimeofday(&tim, 0);
  return (tim.tv_sec+(tim.tv_usec/1000000.0)) - m_zero;
#else // WINDOWS
    static LARGE_INTEGER t = {0};
    QueryPerformanceCounter(&t);
    return (double)(t.QuadPart-m_zero.QuadPart)/m_freq.QuadPart;
#endif
}

/*! \brief Get time difference.
 *
 *  Reads the current time and return the difference since the last
 *  time this method was invoked. The first time it returns the time
 *  since the class was initilized or the last time ResetTime was invoked.
 *
 *  \return Time difference in seconds.
 */
double Timer::GetDeltaTime()
{
    double t2 = GetTime();
    double dt = t2 - m_time;
    m_time = t2;
    return dt;
}

/*! \brief Get time difference with a lower limit by max fps.
 *
 *  Reads the current time and return the difference since the last
 *  time this method was invoked, but if the deltatime is lower than
 *  1/max_fps it will wait until that time is reached. The first time 
 *  it uses the time since the class was initilized or the last time 
 *  ResetTime was invoked as reference.
 *
 *  \param max_fps Maximal number of frames per second.
 *  \return Time difference in seconds.
 */
double Timer::GetDeltaTime(double max_fps)
{
    double t2 = GetTime();
    double dt = t2 - m_time;
    double wait = 1/max_fps - dt;
    
    if(wait > 0.01)
    {
#ifdef WIN32
        Sleep((DWORD)(1000*wait-10));
#else
        usleep((long)(1000000.0*wait)-10000);
#endif
    }

    t2 = GetTime();
    dt = t2 - m_time;
    wait = 1/max_fps - dt;
  
    while(wait > 0)
    {
#ifdef WIN32
        Sleep(0);
#else
        usleep(0);
#endif
        t2 = GetTime();
        dt = t2 - m_time;
        wait = 1/max_fps - dt;
    }

    m_time = t2;
    return dt;
}

/*! \brief Reset time zero position.
 *
 *  Set reference time to current time or counter value to be used when
 *  calculating time differences. This method is called from the constructor.
 */
void Timer::ResetTime()
{
#ifdef WIN32
    QueryPerformanceCounter(&m_zero);
#else
    m_zero = GetTime();
#endif
}
