/*! \file vec3.h
 *  \brief Vector file.
 *
 *  This file contains a vector type with most of the common
 *  operators implemented.
 */

#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <cmath>
#include <iostream>

/*! \brief Vector type for 3D.
 *
 *  Vector type for 3D with most of the common
 *  operators included.
 */
template< typename T >
struct Vec3
{
    T x, y, z; /*!< Component values */

    /*! \brief Constructor
     *
     *  Sets the components of the vector to 0.
     */
    inline Vec3()
    { x=0; y=0; z=0; }

    /*! \brief Constructor
     *
     *  Sets the components of the vector.
     *
     *  \param X x-value.
     *  \param Y y-value.
     *  \param Z z-value.
     */
    inline Vec3( const T X, const T Y, const T Z )
    { x=X; y=Y; z=Z; }

    /*! \brief Constructor
     *
     *  Sets the components of the vector to a scalar.
     *
     *  \param A value for all components of the vector.
     */    
    inline Vec3( const T A )
    { x=A; y=A; z=A; }

    /*! \brief Constructor
     *
     *  Sets the vector to another vector.
     *
     *  \param v vector to initialize to.
     */
    inline Vec3( const Vec3<T>& v )
    { x=v.x; y=v.y; z=v.z; }

    /*! \brief operator []
     *
     *  Access the elements of the vector using the [] operator.
     *
     *  \param i number of component.
     *  \return scalar of type T.
     */
    inline T operator [] (const int i) const
    { return ((T*)this)[i]; }

    /*! \brief operator *
     *
     *  Access the elements of the vecotr as a pointer.
     *
     *  \return pointer to scalar of type T.
     */
    inline T* operator * ()
    { return (T*)this; }

    /*! \brief Negate vector
     *
     *  Return the vector negated.
     *
     *  \return negated vector.
     */
    inline Vec3<T> operator - () const
    { return Vec3<T>(-x, -y, -z); }

    /*! \brief Assignment operator.
     * 
     *  Assign a vector to another.
     *
     *  \param the vector to copy values from.
     *  \return the changed vector.
     */
    inline Vec3<T>& operator = ( const Vec3<T>& v )
    { x=v.x; y=v.y; z=v.z; return *this; }

    /*! \brief Add and assign operator.
     * 
     *  Add a vector to another.
     *
     *  \param v vector to add from.
     *  \return the changed vector.
     */
    inline Vec3<T>& operator += ( const Vec3<T> &v )
    { x=x+v.x; y=y+v.y; z=z+v.z; return *this; }

    /*! \brief Subtract and assign operator.
     *
     *  Subtract a vector from another.
     *
     *  \param v vector to subtract from.
     *  \return the changed vector.
     */
    inline Vec3<T>& operator -= ( const Vec3<T> &v )
    { x=x-v.x; y=y-v.y; z=z-v.z; return *this; }

    /*! \brief Clamp values in vector.
     *
     *  Clamp the values in the vector between low and high.
     *
     *  \param low lower limit for values.
     *  \param high high limit for values.
     *  \return the changed vector.
     */
    inline Vec3<T>& Clamp( const T& low,  const T& high )
    {
      x = x<low ? low : (x>high ? high : x);
      y = y<low ? low : (y>high ? high : y);
      z = z<low ? low : (z>high ? high : z);
      return *this;
    }

    /*! \brief Cross product
     *
     *  Outer product between two vectors.
     *
     *  \param v right vector in the cross product.
     *  \return cross product between the vectors.
     */
    inline Vec3<T> Cross( const Vec3<T>& v ) const
    { return Vec3<T>(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x); }

    /*! \brief Dot product
     *
     *  Inner product between two vectors.
     *
     *  \param v vector to do dot product with.
     *  \return dotproduct between the vectors.
     */
    inline T Dot( const Vec3<T>& v ) const
    { return x*v.x+y*v.y+z*v.z; }

    /*! \brief Norm of the vector.
     *
     *  Calculate the length of the vector.
     *
     *  \return length of the vector.
     */
    inline T Norm() const
    { return (T)std::sqrt(x*x+y*y+z*z); }

    /*! \brief Normalize the vector.
     *
     *  Normalize the vector.
     *
     *  \return the normalized vector.
     */
    inline Vec3<T>& Normalize()
    {
        float norm=Norm();
        x = x/norm;
        y = y/norm;
        z = z/norm;
        return *this; 
    }
};

/*! \brief Output operator for vector class.
 *
 *  Outputs a vector to an output stream stream.
 *
 *  \param os output stream.
 *  \param v vector.
 *  \return output stream.
 */
template< typename T >
std::ostream& operator << (std::ostream& os, const Vec3<T>& v)
{
    return os<<"["<<v.x<<", "<<v.y<<", "<<v.z<<"]";
}

/*! \brief Compare operator for vector class.
 *
 *  Compare two vectors.
 *
 *  \param a left vector.
 *  \param a right vector.
 *  \return result of the comparsion.
 */
template< typename T >
bool operator < (const Vec3<T>& a, const Vec3<T>& b)
{
    return (a.x*a.x+a.y*a.y+a.z*a.z) < (b.x*b.x+b.y*b.y+b.z*b.z);
}

/*! \brief Compare operator for vector class.
 *
 *  Compare two vectors.
 *
 *  \param a left vector.
 *  \param a right vector.
 *  \return result of the comparsion.
 */
template< typename T >
bool operator <= (const Vec3<T>& a, const Vec3<T>& b)
{
    return (a.x*a.x+a.y*a.y+a.z*a.z) < (b.x*b.x+b.y*b.y+b.z*b.z);
}

/*! \brief Compare operator for vector class.
 *
 *  Compare two vectors.
 *
 *  \param a left vector.
 *  \param a right vector.
 *  \return result of the comparsion.
 */
template< typename T >
bool operator == (const Vec3<T>& a, const Vec3<T>& b)
{
    return (a.x==b.x && a.y==b.y && a.z==b.z);
}

/*! \brief Compare operator for vector class.
 *
 *  Compare two vectors.
 *
 *  \param a left vector.
 *  \param a right vector.
 *  \return result of the comparsion.
 */
template< typename T >
bool operator > (const Vec3<T>& a, const Vec3<T>& b)
{
    return (a.x*a.x+a.y*a.y+a.z*a.z) > (b.x*b.x+b.y*b.y+b.z*b.z);
}

/*! \brief Compare operator for vector class.
 *
 *  Compare two vectors.
 *
 *  \param a left vector.
 *  \param a right vector.
 *  \return result of the comparsion.
 */
template< typename T >
bool operator >= (const Vec3<T> &a, const Vec3<T> &b)
{
    return (a.x*a.x+a.y*a.y+a.z*a.z) >= (b.x*b.x+b.y*b.y+b.z*b.z);
}

/*! \brief Addition operator.
 *
 *  Addition of two vectors.
 * 
 *  \param a left vector.
 *  \param a right vector.
 *  \return sum of the vectors.
 */
template< typename T >
Vec3<T> operator + (const Vec3<T>& a, const Vec3<T>& b )
{
    return Vec3<T>(a.x+b.x, a.y+b.y, a.z+b.z);
}

/*! \brief Subtraction operator.
 *
 *  Subtraction of two vectors.
 * 
 *  \param a left vector.
 *  \param a right vector.
 *  \return difference between the vectors.
 */
template< typename T >
Vec3<T> operator - (const Vec3<T> &a, const Vec3<T> &b )
{
    return Vec3<T>(a.x-b.x, a.y-b.y, a.z-b.z);
}

/*! \brief Multiplicatoin operator.
 *
 *  Elementwise multiplication of two vectors.
 * 
 *  \param a left vector.
 *  \param a right vector.
 *  \return elementwise product of the vectors.
 */
template< typename T >
Vec3<T> operator * (const Vec3<T> &a, const Vec3<T> &b )
{
    return Vec3<T>(a.x*b.x, a.y*b.y, a.z*b.z);
}

/*! \brief Division operator.
 *
 *  Elementwise division of two vectors.
 * 
 *  \param a left vector.
 *  \param a right vector.
 *  \return elementwise divided vector.
 */
template< typename T >
Vec3<T> operator / (const Vec3<T> &a, const Vec3<T> &b )
{
    return Vec3<T>(a.x/b.x, a.y/b.y, a.z/b.z);
}

/*! \brief Add scalar operator
 *
 *  Add a scalar to each element of the vector.
 *
 *  \param a vector.
 *  \param b scalar.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator + (const Vec3<T>& a, const Arg& b)
{
    return Vec3<T>(a.x+b, a.y+b, a.z+b);
}

/*! \brief Add scalar operator
 *
 *  Add a scalar to each element of the vector.
 *
 *  \param a scalar.
 *  \param b vector.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator + (const Arg& a, const Vec3<T>& b)
{
    return Vec3<T>(a+b.x, a+b.y, a+b.z);
}

/*! \brief Subtract scalar operator
 *
 *  Subtract a scalar from each element of the vector.
 *
 *  \param a vector.
 *  \param b scalar.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator - (const Vec3<T> &a, const Arg& b)
{
    return Vec3<T>(a.x-b, a.y-b, a.z-b);
}

/*! \brief Subtract scalar operator
 *
 *  Add a vector to each element of a negated vector.
 *
 *  \param a scalar.
 *  \param b vector.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator - (const Arg& a, const Vec3<T>& b)
{
    return Vec3<T>(a-b.x, a-b.y, a-b.z);
}

/*! \brief Multiply scalar operator
 *
 *  Multiply a scalar to each element of a vector.
 *
 *  \param a vector.
 *  \param b scalar.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator * (const Vec3<T>& a, const Arg& b)
{
    return Vec3<T>(a.x*b, a.y*b, a.z*b);
}

/*! \brief Multiply scalar operator
 *
 *  Multiply a scalar to each element of a vector.
 *
 *  \param a scalar.
 *  \param b vector.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator * (const Arg& a, const Vec3<T>& b)
{
    return Vec3<T>(a*b.x, a*b.y, a*b.z);
}

/*! \brief Divide by scalar operator
 *
 *  Divide each element of the vector with a scalar.
 *
 *  \param a vector.
 *  \param b scalar.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator / (const Vec3<T>& a, const Arg& b)
{
    return Vec3<T>(a.x/b, a.y/b, a.z/b);
}

/*! \brief Divide a scalar operator
 *
 *  Divide a scalar with each element of the vector.
 *
 *  \param a scalar.
 *  \param b vector.
 *  \return resulting vector.
 */
template< typename T, class Arg >
Vec3<T> operator / (const Arg& a, const Vec3<T> &b)
{
    return Vec3<T>(a/b.x, a/b.y, a/b.z);
}

/*! \brief Vector of floats
 *
 *  Predefined vector type with floats as elements.
 */
typedef Vec3<float> Vec3f;

/*! \brief Vector of integers
 *
 *  Predefined vector type with integers as elements.
 */
typedef Vec3<int> Vec3i;

#endif // End of _VECTOR_H_
